{----------------------------------------------------------------------------
|
| Library: Spider Containers for Object Pascal
|
| Module: ObjList.Pas
|
| Description: List containers.
|
| Copyright (c) 1996,97 Michel Brazeau, Interval Software
|
| History: Version 1.0  March 1996. 
|
|---------------------------------------------------------------------------}
unit ObjList;

{$I Contain.Inc}

interface

uses
    ObjPerst, { TObjectReader, TObjectWriter }
    ObjCmp,   { TCompareFunc }
    ObjBase;  { TNodeContainer }

type

{ doubly-linked list node }
TListNode = class(TNode)
public
     Prev,
     Next    : TListNode;

     constructor Create;

     destructor Destroy; override;
end;

type

{ Ordered doubly link list. The list is maintained in a ordered fashion
  on every insertion. This requires some overhead, large lists (more than
  500 to 2000) items should use the TBinaryTree which offers a better
  performance for insertion and searches, at the expense of using more
  memory }
TOrderedList = class(TNodeContainer)
public
    constructor Create( const InObjectType  : TClass;
                        const InCompareFunc : TCompareFunc
                       );

    destructor Destroy; override;

    { inserts a node at the proper position according to the compare
      function. The current position is set to the node just inserted. }
    procedure Insert(const Obj : TObject); override;

    { Deletes (removes and Frees) the N'th matching Occur'ence of Obj.
      Returns True if the object is found and deleted. Positions the
      current node to the next node if possible }
    function Delete( const Obj   : TObject;
                     const Occur : LongInt
                    ) : Boolean; override;

    { deletes the current node, without freeing the object in the node.
      It is then the responsibility of the user to retrieve the object first,
      in order to later free the object. Current is set to the next node
      if possible }
    function DetachCurrent : TObject; override;

    { searches for the N'th matching Occur'ence of Obj, and if found, the
      current node is set to it. If not found, the current position is
      not changed }
    function Search( const Obj   : TObject;
                     const Occur : LongInt
                    ) : Boolean; override;

    { GotoFirst, GotoLast return False when the list is emtpy.
      EContainerIsEmpty is not raised. }
    function GotoFirst : Boolean; override;
    function GotoLast  : Boolean; override;

    { returns False if list is empty or last node is reached }
    function GotoNext : Boolean; override;

    { returns False if list is empty or first node is reached }
    function GotoPrev : Boolean; override;

    { for each item in the container, call ProcIterator. ProcIterator is
      a file level procedure. }
    procedure ForEachCallProc( const ProcIterator : TProcIterator;
                               const CallBackData : LongInt ); override;

    { for each item in the container, call MethodIterator. MethodIterator is
      a method of a class. }
    procedure ForEachCallMethod( const MethodIterator : TMethodIterator;
                                 const CallBackData   : LongInt ); override;

    procedure Clear; override;

protected

    Tail,                  { pointer to the Tail of list }
    Head     : TListNode;  { pointer to Head of list     }

    function GetNthObj(const Index : LongInt) : TObject; override;

    { searches for the N'th matching Occur'ence of Obj, and if found, the
      current node is set to it. If not found, the current node is
      positioned on the first node greater than Obj, or is nil, if all
      objects in the list are smaller than Obj or the list is empty) }
    function SearchAndPosition( const Obj     : TObject;
                                const Occur         : LongInt
                              ) : Boolean;

end;

type

{ Unordered doubly link list. Insertions are very quick since no positioning
  needs to be perform to insert the node }
TUnorderedList = class(TOrderedList)
public
    { When assigning an object the current node, the previous object is
      Free'd. }
    property CurrentObj : TObject read  GetCurrentObj
                                  write SetCurrentObj;

    { inserts a node after the current node. The current position is set to
      the node just inserted. }
    procedure Insert(const Obj : TObject); override;

    { insert a node at Head or Tail }
    procedure InsertAtHead(const Obj : TObject);
    procedure InsertAtTail(const Obj : TObject);

    { same description TOrderedList }
    function Search( const Obj     : TObject;
                     const Occur   : LongInt
                    ) : Boolean; override;

end;

{ Circular list. This list is identical to the Unordered list, but calling
  GotoNext at the end of list, moves to the first object. Similarily,
  calling, GotoPrev at the beginning of the list, moves to the last object }
TCircularList = class(TUnOrderedList)
public
    { returns False only if list is empty. If at end of list, CurrentObj
      is set to the beginning of list  }
    function GotoNext : Boolean; override;

    { returns False only if list is empty. If at beginning of list, CurrentObj
      is set to the end of list  }
    function GotoPrev : Boolean; override;

end;


{ stream registration routines }
procedure ReadList( const Obj    : TObject;
                    const Reader : TObjectReader ); far;

procedure WriteList( const Obj    : TObject;
                     const Writer : TObjectWriter ); far;

{--------------------------------------------------------------------}

implementation

uses
    ObjType;   { RaiseXXX }

{ TListNode implementation -------------------------------------------}

constructor TListNode.Create;
begin
    inherited Create;
end;

destructor TListNode.Destroy;
begin
    inherited Destroy;
end;

{ TOrderedist implementation --------------------------------------}

constructor TOrderedList.Create( const InObjectType   : TClass;
                                 const InCompareFunc  : TCompareFunc
                                );
begin
    inherited Create(InObjectType, InCompareFunc);

    Head        := nil;
    Tail        := nil;
end;

{--------------------------------------------------------------------}

destructor TOrderedList.Destroy;
begin
    { inherited Destroy calls Clear to empty the list }
    inherited Destroy;
end;

{--------------------------------------------------------------------}

procedure TOrderedList.Insert(const Obj : TObject);

var
    Node : TListNode;

begin
    {$ifndef __RemoveTypeSafety}
    CheckType(Obj);
    {$endif}

    SearchAndPosition(Obj, 1);

    { assert : Current is at the positioned > Obj, or is nil (if all
      objects are smaller or list is empty) }

    if Current = nil then
    begin
        { item must be inserted at tail }
        Current     := TListNode.Create;
        Current.Obj := Obj;

        if Head <> nil then
        begin
            Tail.Next                   := (Current as TListNode);
            (Current as TListNode).Prev := Tail;
            (Current as TListNode).Next := nil;
            Tail                        := (Current as TListNode);
        end
        else
        begin
            { list was empty }
            Head      := (Current as TListNode);
            Head.Next := nil;
            Head.Prev := nil;
            Tail      := (Current as TListNode);
        end;
    end
    else if (Current as TListNode).Prev = nil then
    begin
        { item must be inserted at Head }
        Current                     := TListNode.Create;
        (Current as TListNode).Next := Head;
        (Current as TListNode).Prev := nil;
        Head.Prev                   := (Current as TListNode);
        Current.Obj                 := Obj;
        Head                        := (Current as TListNode);
    end
    else
    begin
        { item inserted between first and last }
        Node := TListNode.Create;

        Node.Next                        := (Current as TListNode);
        Node.Prev                        := (Current as TListNode).Prev;
        (Current as TListNode).Prev.Next := Node;
        (Current as TListNode).Prev      := Node;
        Node.Obj                         := Obj;
        Current                          := Node;
    end;

    Inc(FSize);
end;

{------------------------------------------------------------------------}

function TOrderedList.GetNthObj(const Index : LongInt) : TObject;
var
    I     : LongInt;

begin
    { check bounds on N }
    if (Index<1) or (Index>Size) then
        RaiseInvalidIndex;

    if Index < (Size div 2) then
    begin
	GotoFirst;

	for I := 2 to Index do
	    GotoNext;
    end
    else
    begin
	GotoLast;

	for I := 1 to (Size-Index) do
	    GotoPrev;
    end;

    Result := CurrentObj;
end;

{------------------------------------------------------------------------}

function TOrderedList.SearchAndPosition( const Obj     : TObject;
                                         const Occur   : LongInt
                                       ) : Boolean;
var
    Comparison   : TComparison;

    CurOccur     : LongInt;  { current occurence }

begin
    {$ifndef __RemoveTypeSafety}
    CheckType(Obj);
    {$endif}

    if Occur <= 0 then
        RaiseInvalidOccurrenceSought;

    Comparison := cmpLess;   { prevent compiler warning }

    CurOccur   := Occur;

    Current    := Head;
    while (Current <> nil) and (CurOccur > 0) do
    begin
        Comparison := FCompareFunc(Obj, Current.Obj);

        if (Comparison = cmpEqual) or
           (Comparison = cmpLess) then
            Dec(CurOccur);

        if CurOccur > 0 then
            Current := (Current as TListNode).Next;
    end;

    Result := (Comparison = cmpEqual) and
              (CurOccur = 0);          

end;

{------------------------------------------------------------------------}

function TOrderedList.Search( const Obj     : TObject;
                              const Occur   : LongInt
                             ) : Boolean;
var
    SavedCurrent : TListNode;
begin
    SavedCurrent := (Current as TListNode);

    Result := SearchAndPosition(Obj, Occur);


    { Current may be nil after SearchAndPosition, so make sure the
      previous position is restored }
    if not Result then
        Current := SavedCurrent;
end;

{--------------------------------------------------------------------}

function TOrderedList.Delete( const Obj     : TObject;
                              const Occur   : LongInt
                             ) : Boolean;

begin
    {$ifndef __RemoveTypeSafety}
    CheckType(Obj);
    {$endif}

    Result := Search(Obj, Occur);

    if Result then
        { assert : Current points to the found record }
        DeleteCurrent;
end;

{--------------------------------------------------------------------}

function TOrderedList.DetachCurrent : TObject;
var
    Node : TListNode;
begin
    if Current = nil then
        RaiseContainerIsEmpty;

    Result := Current.Obj;

    { set Current.Obj to nil, to prevent it from being destroyed, when the
      node is destroyed }
    Current.Obj := nil;

    if Size = 1 then
    begin
        { lone element }
        Current.Free;
        Current := nil;
        Head    := nil;
        Tail    := nil
    end
    else if (Current as TListNode).Next = nil then
    begin
        { last element }
        Tail      := (Current as TListNode).Prev;
        Tail.Next := nil;
        Current.Free;
        Current := Tail;
    end
    else if (Current as TListNode).Prev = nil then
    begin
        { first element }
        Head       := (Current as TListNode).Next;
        Head.Prev  := nil;
        Current.Free;
        Current := Head;
    end
    else
    begin
        { element found between first and last }
        (Current as TListNode).Prev.Next := (Current as TListNode).Next;
        (Current as TListNode).Next.Prev := (Current as TListNode).Prev;
        Node                             := (Current as TListNode).Next;
        Current.Free;
        Current := Node;
    end;

    Dec(FSize);

end;

{------------------------------------------------------------------------}

function TOrderedList.GotoFirst : Boolean;
begin
    Current := Head;

    Result  := Current <> nil;
end;


{------------------------------------------------------------------------}

function TOrderedList.GotoLast : Boolean;
begin
    Current := Tail;

    Result := Current <> nil;
end;

{------------------------------------------------------------------------}

function TOrderedList.GotoNext : Boolean;
begin
    Result := False;

    if Current <> Tail then
    begin
        Current := (Current as TListNode).Next;

	Result  := True;
    end;
end;

{------------------------------------------------------------------------}

function TOrderedList.GotoPrev : Boolean;
begin
    Result := False;

    if Current <> Head then
    begin
        Current := (Current as TListNode).Prev;

	Result   := True;
    end;
end;

{-----------------------------------------------------------------------}

procedure TOrderedList.ForEachCallProc( const ProcIterator : TProcIterator;
                                        const CallBackData : LongInt );
begin
    if GotoFirst then
    repeat
        ProcIterator(CurrentObj, CallBackData);
    until not GotoNext;
end;

{-----------------------------------------------------------------------}

procedure TOrderedList.ForEachCallMethod( const MethodIterator : TMethodIterator;
                                          const CallBackData   : LongInt);
begin
    if GotoFirst then
    repeat
        MethodIterator(CurrentObj, CallBackData);
    until not GotoNext;
end;

{-----------------------------------------------------------------------}

procedure TOrderedList.Clear;
begin
    while Head <> nil do
    begin
        Current := Head;
        Head    := Head.Next;

        ReleaseObj(Current.Obj);

        { set Current.Obj to nil so it is not destroyed twice }
        Current.Obj := nil;

        Current.Free;
    end;

    FSize    := 0;

    Tail    := nil;
    Current := nil;
end;

{ TUnorderedList implementation----------------------------------------}

function TUnorderedList.Search( const Obj     : TObject;
                                const Occur   : LongInt
                               ) : Boolean;

var
    Comparison   : TComparison;

    SavedCurrent : TListNode;

    CurOccur     : LongInt;

begin
    {$ifndef __RemoveTypeSafety}
    CheckType(Obj);
    {$endif}

    if Occur <= 0 then
        RaiseInvalidOccurrenceSought;

    SavedCurrent := (Current as TListNode);

    CurOccur := Occur;

    Current := Head;
    while (Current <> nil) and (CurOccur > 0) do
    begin
        Comparison := FCompareFunc(Obj, Current.Obj);

        if Comparison = cmpEqual then
            Dec(CurOccur);

        if CurOccur > 0 then
            Current := (Current as TListNode).Next;
    end;

    Result := (CurOccur = 0);

    if not Result then
        Current := SavedCurrent;
end;

{---------------------------------------------------------------------}

procedure TUnorderedList.Insert(const Obj : TObject);
var
    Node : TListNode;

begin
    {$ifndef __RemoveTypeSafety}
    CheckType(Obj);
    {$endif}

    Node      := TListNode.Create;
    Node.Obj  := Obj;

    Node.Prev := nil;
    Node.Next := nil;

    if (Head = nil) then
    begin
        { inserting when the list is empty }
	Head    := Node;
	Tail    := Node;
	Current := Node;
    end
    else if Current = Tail then
    begin
        { insert at end of list }
	Tail.Next := Node;
	Node.Prev := Tail;
	Tail      := Node;
	Current   := Node;
    end
    else
    begin
        { insert node after the current node }
        (Current as TListNode).Next.Prev := Node;
        Node.Next                        := (Current as TListNode).Next;

	Node.Prev                   := (Current as TListNode);
	(Current as TListNode).Next := Node;

	Current := Node;
    end;

    Inc(FSize);
end;

{---------------------------------------------------------------------}

procedure TUnorderedList.InsertAtHead(const Obj : TObject);
begin
    {$ifndef __RemoveTypeSafety}
    CheckType(Obj);
    {$endif}

    Current                     := TListNode.Create;
    (Current as TListNode).Prev := nil;
    (Current as TListNode).Next := Head;
    Current.Obj                 := Obj;

    if (Tail = nil) then
    begin
        { previously empty }
        Tail := (Current as TListNode);
        Head := Tail;
    end
    else
    begin
        Head.Prev := (Current as TListNode);
        Head      := (Current as TListNode);
    end;

    Inc(FSize);
end;

{---------------------------------------------------------------------}

procedure TUnorderedList.InsertAtTail(const Obj : TObject);
begin
    {$ifndef __RemoveTypeSafety}
    CheckType(Obj);
    {$endif}

    Current                     := TListNode.Create;
    (Current as TListNode).Next := nil;
    (Current as TListNode).Prev := Tail;
    Current.Obj := Obj;

    if (Head = nil) then
    begin
        { previously empty }
        Head := (Current as TListNode);
        Tail := Head;
    end
    else
    begin
        Tail.Next := (Current as TListNode);
        Tail      := (Current as TListNode);
    end;

    Inc(FSize);
end;

{ TCircularList  implentation --------------------------------------------}

function TCircularList.GotoNext : Boolean;
begin
    if Size = 0 then
        Result := False
    else
    begin
        if not inherited GotoNext then
            GotoFirst;

        Result := True;
    end;
end;

{---------------------------------------------------------------------}

function TCircularList.GotoPrev : Boolean;
begin
    if Size = 0 then
        Result := False
    else
    begin
        if not inherited GotoPrev then
            GotoLast;

        Result := True;
    end;
end;

{---------------------------------------------------------------------}

procedure ReadList( const Obj    : TObject;
                    const Reader : TObjectReader );
begin
    { all lists are derived from TOrderedList. Be careful if this is not the
      case in future versions. }
    with Obj as TOrderedList do
    begin
        Tail := nil;
        Head := nil;
    end;

    ReadNodeContainer( Obj, Reader );
end;

{---------------------------------------------------------------------}

procedure WriteList( const Obj    : TObject;
                     const Writer : TObjectWriter );
begin
    WriteNodeContainer( Obj, Writer );
end;

{---------------------------------------------------------------------}

initialization

    {$ifdef Win32}
    RegisterStreamable( TOrderedList,
                        [TObjectReadProc(@ReadList)],
                        [TObjectWriteProc(@WriteList)] );

    RegisterStreamable( TUnOrderedList,
                        [TObjectReadProc(@ReadList)],
                        [TObjectWriteProc(@WriteList)] );

    RegisterStreamable( TCircularList,
                        [TObjectReadProc(@ReadList)],
                        [TObjectWriteProc(@WriteList)] );
    {$else}
    RegisterStreamable( TOrderedList,
                        [ReadList],
                        [WriteList] );

    RegisterStreamable( TUnOrderedList,
                        [ReadList],
                        [WriteList] );

    RegisterStreamable( TCircularList,
                        [ReadList],
                        [WriteList] );
    {$endif}

end.



