{----------------------------------------------------------------------------
|
| Library: Spider Containers for Object Pascal
|
| Module: ObjQueue.Pas
|
| Description: Queue and priority queue containers.
|
| Copyright (c) 1996,97 Michel Brazeau, Interval Software
|
| History: Version 1.0  March 1996. 
|
|---------------------------------------------------------------------------}
unit ObjQueue;

{$I Contain.Inc}

interface

uses
    ObjPerst, { TObjectReader, TObjectWriter }
    ObjCmp,   { TCompareFunc, CompareEqual }
    ObjBase;  { TContainer, TNodeContainer }

type

{ First In/First Out data structure class }
TQueue = class(TContainer)
protected
    { embedded container }
    Container        : TNodeContainer;

    function GetSize : LongInt; override;
    function QueryEmpty : Boolean; override;

    procedure SetOwnership( const InOwnership : Boolean ); override;

public
    constructor Create(const InObjectType : TClass);

    { destroys calls Clear to remove and free every element in the stack }
    destructor Destroy; override;

    { clear all the items in the stack. Free is called for each item }
    procedure Clear; override;

    { add an element to the queue }
    procedure EnQueue(const Obj : TObject); virtual;

    { removes the object that has been waiting in the queue for the longest
      time }
    function DeQueue : TObject; virtual;

    { for each item in the container, call ProcIterator. ProcIterator is
      a file level procedure. }
    procedure ForEachCallProc( const ProcIterator : TProcIterator;
                               const CallBackData   : LongInt ); override;

    { for each item in the container, call MethodIterator. MethodIterator is
      a method of a class. }
    procedure ForEachCallMethod( const MethodIterator : TMethodIterator;
                                 const CallBackData   : LongInt); override;

end;

{ Priority queue. This is a queue where elements with the highest priority
  (a lower value implies a higher priority, determined by the compare
  function) are always removed from the queue first. When elements in the
  queue have the same priority, they are treated in a first come first serve
  basis }
TPriorityQueue = class(TQueue)

    constructor Create( const InObjectType  : TClass;
                        const InCompareFunc : TCompareFunc
                       );

    procedure EnQueue(const Obj : TObject); override;

    function DeQueue : TObject; override;
end;

{ stream registration routines. ReadQueue and WriteQueue are also valid for
  TPriorityQueue class. }
procedure ReadQueue( const Obj    : TObject;
                     const Reader : TObjectReader ); far;

procedure WriteQueue( const Obj    : TObject;
                      const Writer : TObjectWriter ); far;


{--------------------------------------------------------------------------}

implementation

uses
    ObjType,  { RaiseXXX }
    ObjBTree, { TBinaryTree }
    ObjList;  { TUnOrderedList }

{--------------------------------------------------------------------------}

constructor TQueue.Create(const InObjectType : TClass);
begin
    inherited Create(InObjectType);

    Container := TUnOrderedList.Create(InObjectType, CompareEqual);

    { MB May 13, 1997. Set the base class FCompareFunc for streaming purposes }
    FCompareFunc := CompareEqual;
end;

{--------------------------------------------------------------------------}

destructor TQueue.Destroy;
begin
    { called inherited Destroy first, so when Clear is called, Container is
      still valid }
    inherited Destroy;

    Container.Free;
end;

{--------------------------------------------------------------------------}

procedure TQueue.SetOwnership( const InOwnership : Boolean );
begin
    { maintain the FOwnership data member in the base class }
    inherited SetOwnership(InOwnership);

    Container.Ownership := InOwnership;
end;

{--------------------------------------------------------------------------}

procedure TQueue.Clear;
begin
    { check for nil, as Clear may be invoked on a partially constructed
      object }
    if Container <> nil then
        Container.Clear;
end;

{--------------------------------------------------------------------------}

procedure TQueue.Enqueue(const Obj : TObject);
begin
    (Container as TUnOrderedList).InsertAtTail(Obj);
end;

{--------------------------------------------------------------------------}

function TQueue.DeQueue : TObject;
begin
    if Container.IsEmpty then
        RaiseContainerIsEmpty;

    (Container as TUnOrderedList).GotoFirst;

    { delete the node, but don't free the object in the node, as this is
      the object being returned ! }
    Result := Container.DetachCurrent;
end;

{--------------------------------------------------------------------------}

function TQueue.GetSize : LongInt;
begin
    Result := Container.Size;
end;

{--------------------------------------------------------------------------}

function TQueue.QueryEmpty : Boolean;
begin
    Result := Container.IsEmpty;
end;

{--------------------------------------------------------------------------}

procedure TQueue.ForEachCallProc( const ProcIterator : TProcIterator;
                                  const CallBackData : LongInt);
begin
    Container.ForEachCallProc(ProcIterator, CallBackData);
end;

{--------------------------------------------------------------------------}

procedure TQueue.ForEachCallMethod( const MethodIterator : TMethodIterator;
                                    const CallBackData   : LongInt);
begin
    Container.ForEachCallMethod(MethodIterator, CallBackData);
end;


{ TPriorityQueue implementation ------------------------------------------}

constructor TPriorityQueue.Create( const InObjectType   : TClass;
                                   const InCompareFunc  : TCompareFunc
                                  );
begin
    inherited Create(InObjectType);

    { the priority queue uses a binary tree instead of a list }
    Container.Free;
    Container := TBinaryTree.Create(InObjectType, InCompareFunc);

    { MB May 13, 1997. Set the base class FCompareFunc for streaming purposes }
    FCompareFunc := InCompareFunc;
end;

{--------------------------------------------------------------------------}

procedure TPriorityQueue.EnQueue(const Obj   : TObject);
begin
    (Container as TBinaryTree).Insert(Obj);
end;

{--------------------------------------------------------------------------}

function TPriorityQueue.DeQueue : TObject;
begin
    if Container.IsEmpty then
        RaiseContainerIsEmpty;

    (Container as TBinaryTree).GotoFirst;

    Result := Container.DetachCurrent;
end;

{--------------------------------------------------------------------------}

procedure ReadQueue( const Obj    : TObject;
                     const Reader : TObjectReader );
var
    Queue : TQueue;
begin
    Queue := Obj as TQueue;

    ReadContainer(Obj, Reader);

    Queue.Container := Reader.ReadChildObject as TNodeContainer;
end;

{--------------------------------------------------------------------------}

procedure WriteQueue( const Obj    : TObject;
                      const Writer : TObjectWriter );
var
    Queue : TQueue;
begin
    Queue := Obj as TQueue;

    WriteContainer(Queue, Writer);

    Writer.WriteChildObject( Queue.Container );
end;

{---------------------------------------------------------------------}

initialization

    {$ifdef Win32}
    RegisterStreamable( TQueue,
                        [TObjectReadProc(@ReadQueue)],
                        [TObjectWriteProc(@WriteQueue)] );

    RegisterStreamable( TPriorityQueue,
                        [TObjectReadProc(@ReadQueue)],
                        [TObjectWriteProc(@WriteQueue)] );
    {$else}
    RegisterStreamable( TQueue,
                        [ReadQueue],
                        [WriteQueue] );

    RegisterStreamable( TPriorityQueue,
                        [ReadQueue],
                        [WriteQueue] );
    {$endif}

end.

