{----------------------------------------------------------------------------
|
| Module: ObjType.Pas
|
| Description: Common types regrouped in a unit to prevent circular unit
|              references.
|
| Copyright (c) 1996,1997 Michel Brazeau, Interval Software
|
| History: May 11, 1997. MB first version.
|          Jan 17, 1998. MB Resource strings (in Spider16.Res or Spider32.Res
|                        files) are no longer used for exception messages to
|                        avoid conflicts with other third party libraries. The
|                        strings may be modified at run-time for multilingual
|                        support. See SetSpiderMessages procedure.
|
|---------------------------------------------------------------------------}
unit ObjType;

{$I Contain.Inc}

interface

uses
    SysUtils; { Exception, NewStr, DisposeStr }

{--------------------------------------------------------------------------}

{ Index range for array indexes }
const
    CMaxArrayIndex        = {$ifdef Win32}
                            (MaxLongInt) div 4;
                            {$else}
                            ($FFFF - 16) div 4;
                            {$endif}
    CInvalidArrayIndex    = 0;

type

{ Note: the array indexes are 1 based, but 0 is reserved as an
        invalid array index }
TArrayIndex =  CInvalidArrayIndex..CMaxArrayIndex;

{--------------------------------------------------------------------------}

{ hashing function. Must return an index from 1..TableSize as a function
  of Obj. StringHash, IntegerHash may be used in custom hash
  functions, if hashing is based on strings or integers respectively }
THashFunc = function( const Obj       : TObject;
                      const TableSize : TArrayIndex) : TArrayIndex;


{--------------------------------------------------------------------------}

{ exceptions raised by Spider Container Classes }

type

{ raised when an object passed to container object is not of the proper
  type }
EInvalidObjectType         = class(Exception);

{ Raised when an operation on an empty container is invalid }
EContainerIsEmpty          = class(Exception);

{ raised when an object in a TNodeContainer is not found, due to
  abnormal conditions (not when Search is invoked) }
ENodeNotFound              = class(Exception);

{ raised when an invalid index is referenced on a container }
EInvalidIndex              = class(Exception);

{ raised if a call to GotoMarker is made before a call to SetMarker,
  for containers that support markers }
EMarkerNotSet              = class(Exception);

{ raised when an TArray based container cannot be resized because the
  new size is smaller than the current size }
ENewArraySizeTooSmall      = class(Exception);

{ raised when adding to a container, when its maximum size has been
  reached }
EContainerIsFull           = class(Exception);

{ raised when a binary searched is attempted on an array that is not
  sorted }
EArrayNotSorted            = class(Exception);

{ raised when a negative or zero'th Occurrence of an object is searched
  in a container }
EInvalidOccurrenceSought   = class(Exception);

{ raised when a invalid capacity value ( less then 1 ) is used to
  create a TArray instance }
EInvalidArrayCapacity      = class(Exception);

{ raised when an unregistered object is read from an object stream.
  See ObjPerst.RegisterStreamable }
EClassNotRegistered        = class(Exception);

{ raised when a container is read/written from/to a stream, and the compare
  function has not been registered. ObjPerst.RegisterCompareFunction }
ECompareFuncNotRegistered  = class(Exception);

{ raised when a hash table is read/written from/to a stream, and the hash
  function has not been registered. ObjPerst.RegisterHashFunction }
EHashFuncNotRegistered     = class(Exception);

{ raised when the version of a class is not registered. More specifically,
  a TObjectReadProc is not registered for the version of an object being
  loaded from a stream }
EClassVersionNotRegistered = class(Exception);

{ raised when registering a compare function, and a compare function has
  allready been registered with the same name. }
EDuplicateCompareFunc      = class(Exception);

{ raised when registering a hash function, and a hash function has allready
  been registered with the same name. }
EDuplicateHashFunc         = class(Exception);

{ raised when registering a class, if the class is allready registered. }
EDuplicateRegisteredClass  = class(Exception);


{ procedure wrappers to raise the exceptions }
procedure RaiseInvalidObjectType;
procedure RaiseContainerIsEmpty;
procedure RaiseInvalidIndex;
procedure RaiseMarkerNotSet;
procedure RaiseNewArraySizeTooSmall;
procedure RaiseContainerIsFull;
procedure RaiseArrayNotSorted;
procedure RaiseInvalidOccurrenceSought;
procedure RaiseInvalidArrayCapacity;
procedure RaiseClassNotRegistered(const ClassName : String);
procedure RaiseCompareFuncNotRegistered;
procedure RaiseHashFuncNotRegistered;
procedure RaiseClassVersionNotRegistered;
procedure RaiseDuplicateCompareFunc(const CompareFuncName : String);
procedure RaiseDuplicateHashFunc(const HashFuncName : String);
procedure RaiseDuplicateRegisteredClass(const RegisteredName : String);

{ allows changing the Spider Containers messages. May be used at run-time
  to support multilanguage applications. }
procedure SetSpiderMessages( const InInvalidObjectType          : String;
                             const InContainerIsEmpty           : String;
                             const InContainerIsFull            : String;
                             const InInvalidIndex               : String;
                             const InMarkerNotSet               : String;
                             const InNewArraySizeTooSmall       : String;
                             const InArrayNotSorted             : String;
                             const InInvalidOccurrenceSought    : String;
                             const InInvalidArrayCapacity       : String;
                             const InClassNotRegistered         : String;
                             const InCompareFuncNotRegistered   : String;
                             const InHashFuncNotRegistered      : String;
                             const InClassVersionNotRegistered  : String;
                             const InDuplicateCompareFunc       : String;
                             const InDuplicateHashFunc          : String;
                             const InDuplicateRegisteredClass   : String
                            );


implementation

var
    InvalidObjectType          : String;
    ContainerIsEmpty           : String;
    ContainerIsFull            : String;
    InvalidIndex               : String;
    MarkerNotSet               : String;
    NewArraySizeTooSmall       : String;
    ArrayNotSorted             : String;
    InvalidOccurrenceSought    : String;
    InvalidArrayCapacity       : String;
    ClassNotRegistered         : String;
    CompareFuncNotRegistered   : String;
    HashFuncNotRegistered      : String;
    ClassVersionNotRegistered  : String;
    DuplicateCompareFunc       : String;
    DuplicateHashFunc          : String;
    DuplicateRegisteredClass   : String;


{-------------------------------------------------------------------------}

procedure RaiseInvalidObjectType;
begin
    raise EInvalidObjectType.Create(InvalidObjectType);
end;

{-------------------------------------------------------------------------}

procedure RaiseContainerIsEmpty;
begin
    raise EContainerIsEmpty.Create(ContainerIsEmpty);
end;

{-------------------------------------------------------------------------}

procedure RaiseContainerIsFull;
begin
    raise EContainerIsFull.Create(ContainerIsFull);
end;

{-------------------------------------------------------------------------}

procedure RaiseInvalidIndex;
begin
    raise EInvalidIndex.Create(InvalidIndex);
end;

{-------------------------------------------------------------------------}

procedure RaiseMarkerNotSet;
begin
    raise EMarkerNotSet.Create(MarkerNotSet);
end;

{-------------------------------------------------------------------------}

procedure RaiseNewArraySizeTooSmall;
begin
    raise ENewArraySizeTooSmall.Create(NewArraySizeTooSmall);
end;

{-------------------------------------------------------------------------}

procedure RaiseArrayNotSorted;
begin
    raise EArrayNotSorted.Create(ArrayNotSorted);
end;

{-------------------------------------------------------------------------}

procedure RaiseInvalidOccurrenceSought;
begin
    raise EInvalidOccurrenceSought.Create(InvalidOccurrenceSought);
end;

{-------------------------------------------------------------------------}

procedure RaiseInvalidArrayCapacity;
begin
    raise EInvalidArrayCapacity.Create(InvalidArrayCapacity);
end;

{-------------------------------------------------------------------------}

procedure RaiseClassNotRegistered(const ClassName : String);
begin
    raise EClassNotRegistered.Create(ClassNotRegistered + ' : ' + ClassName);
end;

{-------------------------------------------------------------------------}

procedure RaiseCompareFuncNotRegistered;
begin
    raise ECompareFuncNotRegistered.Create(CompareFuncNotRegistered);
end;

{-------------------------------------------------------------------------}

procedure RaiseHashFuncNotRegistered;
begin
    raise EHashFuncNotRegistered.Create(HashFuncNotRegistered);
end;

{-------------------------------------------------------------------------}

procedure RaiseClassVersionNotRegistered;
begin
    raise EClassVersionNotRegistered.Create(ClassVersionNotRegistered);
end;

{-------------------------------------------------------------------------}

procedure RaiseDuplicateCompareFunc(const CompareFuncName : String);
begin
    raise EDuplicateCompareFunc.Create(DuplicateCompareFunc + ' : ' + CompareFuncName);
end;

{-------------------------------------------------------------------------}

procedure RaiseDuplicateHashFunc(const HashFuncName : String);
begin
    raise EDuplicateHashFunc.Create(DuplicateHashFunc + ' : ' + HashFuncName);
end;

{-------------------------------------------------------------------------}

procedure RaiseDuplicateRegisteredClass(const RegisteredName : String);
begin
    raise EDuplicateRegisteredClass.Create(DuplicateRegisteredClass + ' : ' + RegisteredName);
end;

{-------------------------------------------------------------------------}

procedure SetSpiderMessages( const InInvalidObjectType          : String;
                             const InContainerIsEmpty           : String;
                             const InContainerIsFull            : String;
                             const InInvalidIndex               : String;
                             const InMarkerNotSet               : String;
                             const InNewArraySizeTooSmall       : String;
                             const InArrayNotSorted             : String;
                             const InInvalidOccurrenceSought    : String;
                             const InInvalidArrayCapacity       : String;
                             const InClassNotRegistered         : String;
                             const InCompareFuncNotRegistered   : String;
                             const InHashFuncNotRegistered      : String;
                             const InClassVersionNotRegistered  : String;
                             const InDuplicateCompareFunc       : String;
                             const InDuplicateHashFunc          : String;
                             const InDuplicateRegisteredClass   : String
                            );
begin
    InvalidObjectType          := InInvalidObjectType;
    ContainerIsEmpty           := InContainerIsEmpty;
    ContainerIsFull            := InContainerIsFull;
    InvalidIndex               := InInvalidIndex;
    MarkerNotSet               := InMarkerNotSet;
    NewArraySizeTooSmall       := InNewArraySizeTooSmall;
    ArrayNotSorted             := InArrayNotSorted;
    InvalidOccurrenceSought    := InInvalidOccurrenceSought;
    InvalidArrayCapacity       := InInvalidArrayCapacity;
    ClassNotRegistered         := InClassNotRegistered;
    CompareFuncNotRegistered   := InCompareFuncNotRegistered;
    HashFuncNotRegistered      := InHashFuncNotRegistered;
    ClassVersionNotRegistered  := InClassVersionNotRegistered;
    DuplicateCompareFunc       := InDuplicateCompareFunc;
    DuplicateHashFunc          := InDuplicateHashFunc;
    DuplicateRegisteredClass   := InDuplicateRegisteredClass;
end;

{-------------------------------------------------------------------------}

initialization
    SetSpiderMessages( 'Invalid object type in TContainer',
                       'TContainer is empty',
                       'TContainer is full',
                       'Invalid index in TContainer',
                       'TContainer marker is not set',
                       'TArray new size is too small',
                       'TArray is not sorted',
                       'Invalid occurrence sought in TContainer',
                       'Invalid array capacity in TArray',
                       'Class not registered',
                       'Compare function not registered',
                       'Hash function not registered',
                       'Class version not registered',
                       'Duplicate compare function',
                       'Duplicate hash function',
                       'Duplicate registered class'
                      );
end.



