{----------------------------------------------------------------------
|
| Library: Spider Object Database
|
| Module: OdbIbx.Pas
|
| Description: Object database driver class using Interbase express components.
|
| History: Sep 28, 2000. Michel Brazeau, first version
|          Mar 23, 2001. MB update to latest driver interface, and add
|                        TIbxOdbDriverTest for automated tests.
|
|---------------------------------------------------------------------}
unit OdbIbx;

{$I Odb.Inc}

interface

uses
    {$ifdef __Testing}
    DriverTest, { for TOdbDriverTest }
    {$endif}

    Classes,  { for TMemoryStream }
    SysUtils, { for Exception, StrPas }
    Db,       { for TDatabase, TDataset }
    IBQuery,  { for TIBQuery }
    IBDatabase, { for TIBDatabase }
    ObjList,  { for TUnOrderedList }
    Odb,      { for TObjectFile, TOdbDriver }
    OdbIndex, { for TOdbIndex }
    OdbField; { for TOdbField }

{--------------------------------------------------------------------}

type

TIbxOdbDriver = class(TOdbDriver)
protected

    FDatabase      : TIBDatabase;

    FTransaction   : TIBTransaction;

public

    constructor Create(const InDatabaseName : String); override;
    destructor Destroy; override;

    procedure InternalStartTransaction; override;
    procedure InternalCommit; override;
    procedure InternalRollback; override;

    procedure InternalOpenDatabase(
                            const ServerName : String;
                            const UserName   : String;
                            const Password   : String
                          ); override;

    procedure InternalCloseDatabase; override;

    function InternalOpenDataset( const ObjectDataset : TObjectDataset ) : TOdbHandle; override;

    procedure InternalCloseDataset( const Handle : TOdbHandle); override;

    procedure InternalDeleteDataset( const ObjectDataset : TObjectDataset ); override;

    procedure InternalCreateDataset( const ObjectDataset : TObjectDataset ); override;

    procedure InternalInsertSQL(
                      const Handle      : TOdbHandle;
                      const Stream      : TMemoryStream;
                      const ORMapping   : TORMapping;
                      const DatasetName : String ); override;

    procedure InternalUpdateSQL(
                      const Handle      : TOdbHandle;
                      const Stream      : TMemoryStream;
                      const ORMapping   : TORMapping;
                      const DatasetName : String ); override;

    procedure InternalDeleteSQL( const Handle      : TOdbHandle;
                                 const ORMapping   : TORMapping;
                                 const DatasetName : String ); override;

    function InternalCreateObjectUpdate : TOdbHandle; override;
    procedure InternalDestroyObjectUpdate( const Handle : TOdbHandle ); override;

    procedure InternalGetCurrent(
                          const Handle  : TOdbHandle;
                          const Stream  : TMemoryStream ); override;

    function InternalFindNext( const Handle : TOdbHandle ) : Boolean; override;

    function InternalFindPrior( const Handle : TOdbHandle ) : Boolean; override;

    function InternalFindFirst( const Handle : TOdbHandle ) : Boolean; override;

    function InternalFindLast( const Handle : TOdbHandle ) : Boolean; override;

    function InternalIsEmpty( const Handle : TOdbHandle ) : Boolean; override;

    function GetSupportedFieldTypes : TOdbFieldTypes; override;

    function GetIbxQuery( const Handle : TOdbHandle ) : TIBQuery;

    function GetDataAccessMethods : TDataAccessMethodSet; override;

    function AutoIncSupported              : Boolean; override;
    function IntegerIndexSupported         : Boolean; override;
    function FloatIndexSupported           : Boolean; override;
    function CaseInsensitiveIndexSupported : Boolean; override;
    function DescendingIndexSupported      : Boolean; override;
    function TransactionsSupported         : Boolean; override;
    function FileLockingSupported          : Boolean; override;
    function ObjectLockingSupported        : Boolean; override;
    function SingleIndexOnly               : Boolean; override;

protected
    function IbxOpenQuery( const Name        : String;
                           const ObjectQuery : TObjectQuery ) : TOdbHandle;

    procedure CreateTableWithSQL(
                          const Name          : String;
                          const ObjectDataset : TObjectDataset );

    procedure DeleteTableWithSQL(
                          const Name          : String;
                          const ObjectDataset : TObjectDataset );

end;



{$ifdef __Testing}

type

TIbxOdbDriverTest = class( TOdbDriverTest )
protected
     procedure GetParams( var DriverClass                   : TOdbDriverClass;
                          var DatabaseName                  : String;
                          var ServerName                    : String;
                          var UserName                      : String;
                          var Password                      : String
                        ); override;
end;

{$endif}

{--------------------------------------------------------------------}

implementation

uses
    IBSQL,    { for TIBSQL }
    IBCustomDataset, { for TIBDataset }
    ObjPerst,
    OdbMisc;  { for RaiseOdbException }

const

CObjectFieldName = 'Object';

{--------------------------------------------------------------------}

constructor TIbxOdbDriver.Create(const InDatabaseName : String);
begin
    FDatabase  := TIBDatabase.Create(nil);

    FDatabase.DatabaseName   := InDatabaseName;

    FTransaction             := TIBTransaction.Create(nil);

    FTransaction.AddDatabase(FDatabase);

    inherited Create(InDatabaseName);
end;

{--------------------------------------------------------------------}

destructor TIbxOdbDriver.Destroy;
begin
    { invoke inherited first, to close the database if it is opened before
      invoking FDatabase.Free }
    inherited Destroy;

    FTransaction.Free;

    FDatabase.Free;
end;

{--------------------------------------------------------------------}

procedure TIbxOdbDriver.InternalStartTransaction;
begin
    FTransaction.StartTransaction;
end;

{--------------------------------------------------------------------}

procedure TIbxOdbDriver.InternalCommit;
begin
    FTransaction.Commit;
end;

{--------------------------------------------------------------------}

procedure TIbxOdbDriver.InternalRollback;
begin
    FTransaction.Rollback;
end;

{--------------------------------------------------------------------}

procedure TIbxOdbDriver.InternalOpenDatabase(
                        const ServerName : String;
                        const UserName   : String;
                        const Password   : String
                      );
begin
    FDatabase.LoginPrompt                := False;
    FDatabase.DatabaseName               := ServerName; { Added by AD 31st July 2001}
    FDatabase.Params.Values['USER_NAME'] := UserName;
    FDatabase.Params.Values['PASSWORD']  := Password;

    FDatabase.Connected := True;
end;

{--------------------------------------------------------------------}

procedure TIbxOdbDriver.InternalCloseDatabase;
begin
    FDatabase.Close;
end;

{--------------------------------------------------------------------}

function TIbxOdbDriver.IbxOpenQuery(
                           const Name        : String;
                           const ObjectQuery : TObjectQuery ) : TOdbHandle;
var
    Query       : TIBQuery;
    Transaction : TIBTransaction;

begin
    Query   := nil;
    try
        Query     := TIBQuery.Create(nil);

        { Create a transaction owned by the Query, even if the query
          is readonly. }

        Transaction := TIBTransaction.Create(Query);
        Transaction.AddDatabase(FDatabase);

        Query.Transaction := Transaction;

        Query.Database := FDatabase;
        Query.SQL.Assign( ObjectQuery.Select );

        Transaction.StartTransaction;

        Query.Open;

        Result := NewHandle(Query);
    except
        Query.Free;
        raise;
    end;
end;

{--------------------------------------------------------------------}

function TIbxOdbDriver.InternalOpenDataset(
                       const ObjectDataset : TObjectDataset ) : TOdbHandle;
begin
    if TObjectQuery(ObjectDataset).Select.Count = 0 then
        RaiseOdbException( EOdbSelectStatementNotSpecified,
                           ObjectDataset.DatasetName );

    Result := IbxOpenQuery( ObjectDataset.DatasetName, TObjectQuery(ObjectDataset) );
end;
{--------------------------------------------------------------------}

procedure TIbxOdbDriver.InternalCloseDataset( const Handle : TOdbHandle);
var
    Query : TIBQuery;
begin
    Query := GetIbxQuery(Handle);

    Query.Close;

    ReleaseHandle(Handle);
end;

{--------------------------------------------------------------------}

procedure TIbxOdbDriver.DeleteTableWithSQL(
                          const Name          : String;
                          const ObjectDataset : TObjectDataset );
var
    IBSQL         : TIBSQL;
    InTransaction : Boolean;
begin
    IBSQL       := TIBSQL.Create(nil);
    try
        IBSQL.Database    := FDatabase;                       
        IBSQL.Transaction := FTransaction;

        IBSQL.SQL.Add('drop table ' + ObjectDataset.DatasetName );

        InTransaction := FTransaction.InTransaction;

        if not InTransaction then
            FTransaction.StartTransaction;
        try
            IBSQL.ExecQuery;

            if not InTransaction then
                FTransaction.Commit;
        except
            if not InTransaction then
                FTransaction.Rollback;
            raise;
        end;

    finally
        IBSQL.Free;
    end;
end;

{--------------------------------------------------------------------}

procedure TIbxOdbDriver.InternalDeleteDataset( const ObjectDataset : TObjectDataset );
begin
    DeleteTableWithSQL( ObjectDataset.DatasetName, ObjectDataset );
end;

{--------------------------------------------------------------------}

procedure TIbxOdbDriver.InternalCreateDataset(
                          const ObjectDataset : TObjectDataset
                        );
begin
    CreateTableWithSQL(ObjectDataset.DatasetName, ObjectDataset)
end;

{--------------------------------------------------------------------}

function TIbxOdbDriver.InternalCreateObjectUpdate : TOdbHandle;
begin
    Result := 0;
end;

{--------------------------------------------------------------------}

procedure TIbxOdbDriver.InternalDestroyObjectUpdate( const Handle : TOdbHandle );
begin
end;

{--------------------------------------------------------------------}

procedure TIbxOdbDriver.CreateTableWithSQL(
                          const Name          : String;
                          const ObjectDataset : TObjectDataset );
var
    Statement     : TStringList;
    IBSQL         : TIBSQL;
    LineNo        : Integer;
    InTransaction : Boolean;
begin
    Statement := TStringList.Create;
    IBSQL     := TIBSQL.Create(nil);
    try
        IBSQL.Database    := FDatabase;
        IBSQL.Transaction := FTransaction;

        BuildCreateTableStatement( Statement, ObjectDataset.ORMapping,
                                   ObjectDataset.DatasetName,
                                   'character', 'integer', 'float', 'blob');

        IBSQL.SQL.Assign(Statement);

        InTransaction := FTransaction.InTransaction;

        if not InTransaction then
            FTransaction.StartTransaction;
        try

            IBSQL.ExecQuery;

            BuildCreateIndexStatements( Statement, ObjectDataset );
            for LineNo := 0 to (Statement.Count-1) do
            begin
                IBSQL.SQL.Clear;
                IBSQL.SQL.Add( Statement[LineNo] );
                IBSQL.ExecQuery;
            end;

            if not InTransaction then
                FTransaction.Commit;

        except
            if not InTransaction then
                FTransaction.Rollback;
            raise;
        end;

    finally
        IBSQL.Free;
        Statement.Free;
    end;
end;

{--------------------------------------------------------------------}

{ set the field values in Table given a list of TOdbField's }
procedure SetDatasetFieldValues( const Fields    : TUnOrderedList;
                                 const Dataset   : TIBDataset );
var
    Field : TOdbField;
begin
    if Fields.GotoFirst then
    repeat
        Field := Fields.CurrentObj as TOdbField;

        case Field.FieldType of
            oftString  : Dataset.FieldByName(Field.Name).AsString  := Field.AsString;
            oftInteger : Dataset.FieldByName(Field.Name).AsInteger := Field.AsInteger;
            oftFloat   : Dataset.FieldByName(Field.Name).AsFloat   := Field.AsFloat;
        end;
    until not Fields.GotoNext;
end;

{--------------------------------------------------------------------}

procedure InsertInDataset( const Dataset : TIBDataset;
                           const Fields  : TUnOrderedList;
                           const Stream  : TMemoryStream );
var
    Field   : TBlobField;
begin
    Dataset.Insert;
    try
        SetDatasetFieldValues(Fields, Dataset);

        Field := (Dataset.FieldByName(CObjectFieldName) as TBlobField);

        Field.LoadFromStream(Stream);

        Dataset.Post;

    except
        Dataset.Cancel;
        raise;
    end;
end;

{--------------------------------------------------------------------}

procedure UpdateCurrentInDataset( const Dataset : TIBDataset;
                                  const Fields  : TUnOrderedList;
                                  const Stream  : TMemoryStream );
begin
     try
        SetDatasetFieldValues(Fields, Dataset);
        (Dataset.FieldByName(CObjectFieldName) as TBlobField).LoadFromStream(Stream);
        Dataset.Post;
    except
        Dataset.Cancel;
        raise;
    end;
end;

{--------------------------------------------------------------------}

procedure BuildModifySQL( const SQL         : TStrings;
                          const ORMapping   : TORMapping;
                          const DatasetName : String );
var
    Line  : String;
    Field : TOdbField;
begin
    SQL.Clear;
    SQL.Add('update ' + DatasetName + ' set' );

    Line := '';
    if ORMapping.Fields.GotoFirst then
    repeat
        Field := ORMapping.Fields.CurrentObj as TOdbField;

        if Line = '' then
            Line := Field.Name + ' = :' + Field.Name
        else
            Line := Line + ', ' + Field.Name + ' = :' + Field.Name;
    until not ORMapping.Fields.GotoNext;

    Line := Line + ', object = :object';
    SQL.Add(Line);

    BuildSelectWhereClause( Line, ORMapping, DatasetName );

    SQL.Add(Line);
end;

{--------------------------------------------------------------------}

procedure TIbxOdbDriver.InternalInsertSQL(
                  const Handle      : TOdbHandle;
                  const Stream      : TMemoryStream;
                  const ORMapping   : TORMapping;
                  const DatasetName : String );
var
    IBSQL         : TIBSQL;
    Dataset       : TIBDataset;
    InTransaction : Boolean;

begin
    IBSQL     := TIBSQL.Create(nil);
    Dataset   := TIBDataset.Create(nil);
    try

        IBSQL.Database     := FDatabase;
        IBSQL.Transaction  := FTransaction;
        BuildInsertStatement( IBSQL.SQL, ORMapping.Fields, DatasetName );

        Dataset.Database    := FDatabase;
        Dataset.Transaction := FTransaction;
        BuildSelectObjectQuery( Dataset.SelectSQL, ORMapping, DatasetName );

        BuildModifySQL(Dataset.ModifySQL, ORMapping, DatasetName);

        InTransaction := FTransaction.InTransaction;
        if not InTransaction then
            FTransaction.StartTransaction;
        try
            { first insert the record without the blob }
            IBSQL.ExecQuery;

            { use the TIBDataset to update the blob }
            Dataset.Open;
            Dataset.Edit;
            UpdateCurrentInDataset( Dataset, ORMapping.Fields, Stream );

            if not InTransaction then
                FTransaction.Commit;
        except
            if not InTransaction then
                FTransaction.Rollback;
            raise;
        end;

    finally
        Dataset.Free;
        IBSQL.Free;
    end;
end;

{--------------------------------------------------------------------}

procedure TIbxOdbDriver.InternalUpdateSQL(
                      const Handle      : TOdbHandle;
                      const Stream      : TMemoryStream;
                      const ORMapping   : TORMapping;
                      const DatasetName : String );
var
    Dataset       : TIBDataset;
    InTransaction : Boolean;

begin
    Dataset   := TIBDataset.Create(nil);
    try
        Dataset.Database    := FDatabase;
        Dataset.Transaction := FTransaction;
        BuildSelectObjectQuery( Dataset.SelectSQL, ORMapping, DatasetName );

        BuildModifySQL(Dataset.ModifySQL, ORMapping, DatasetName);

        InTransaction := FTransaction.InTransaction;
        if not InTransaction then
            FTransaction.StartTransaction;
        try
            { Dataset.Open will locate the record to update }
            Dataset.Open;

            if Dataset.Bof and Dataset.Eof then
                RaiseOdbException(EOdbObjectNotFound, 'TIbxOdbDriver.InternalUpdateSQL');

            Dataset.Edit;
            UpdateCurrentInDataset( Dataset, ORMapping.Fields, Stream );

            if not InTransaction then
                FTransaction.Commit;
        except
            if not InTransaction then
                FTransaction.Rollback;
            raise;
        end;

    finally
        Dataset.Free;
    end;
end;

{--------------------------------------------------------------------}
                
procedure TIbxOdbDriver.InternalDeleteSQL(
                             const Handle      : TOdbHandle;
                             const ORMapping   : TORMapping;
                             const DatasetName : String );
var
    IBSQL         : TIBSQL;
    InTransaction : Boolean;

begin
    { Delete the record using the related edit query. }
    IBSQL   := TIBSQL.Create(nil);
    try
        IBSQL.Database     := FDatabase;
        IBSQL.Transaction  := FTransaction;

        BuildDeleteStatement( IBSQL.SQL, ORMapping, DatasetName );

        InTransaction := FTransaction.InTransaction;
        if not InTransaction then
            FTransaction.StartTransaction;
        try
            IBSQL.ExecQuery;

            if not InTransaction then
                FTransaction.Commit;
        except
            if not InTransaction then
                FTransaction.Rollback;
            raise;
        end;

    finally
        IBSQL.Free;
    end;
end;

{--------------------------------------------------------------------}

procedure TIbxOdbDriver.InternalGetCurrent(
                      const Handle  : TOdbHandle;
                      const Stream  : TMemoryStream );
var
    Query   : TIBQuery;
    Field   : TBlobField;
begin
    Query := GetIbxQuery(Handle);

    Field := (Query.FieldByName(CObjectFieldName) as TBlobField);

    Field.SaveToStream(Stream);
end;

{--------------------------------------------------------------------}

function TIbxOdbDriver.InternalFindNext( const Handle : TOdbHandle ) : Boolean;
var
    Query : TIBQuery;
begin
    Query := GetIbxQuery(Handle);
    Query.Next;
    Result := not Query.EOF;
end;

{--------------------------------------------------------------------}

function TIbxOdbDriver.InternalFindPrior( const Handle : TOdbHandle ) : Boolean;
var
    Query : TIBQuery;
begin
    Query := GetIbxQuery(Handle);
    Query.Prior;
    Result := not Query.BOF;
end;

{--------------------------------------------------------------------}

function TIbxOdbDriver.InternalFindFirst( const Handle : TOdbHandle ) : Boolean;
var
    Query : TIBQuery;
begin
    Query := GetIbxQuery(Handle);
    Query.First;
    Result := not (Query.EOF and Query.BOF);
end;

{--------------------------------------------------------------------}

function TIbxOdbDriver.InternalFindLast( const Handle : TOdbHandle ) : Boolean;
var
    Query : TIBQuery;
begin
    Query := GetIbxQuery(Handle);
    Query.Last;
    Result := not (Query.EOF and Query.BOF);
end;

{--------------------------------------------------------------------}

function TIbxOdbDriver.InternalIsEmpty( const Handle : TOdbHandle ) : Boolean;
var
    Query : TIBQuery;
begin
    Query  := GetIbxQuery(Handle);
    Result := Query.BOF and Query.EOF;
end;

{--------------------------------------------------------------------}

function TIbxOdbDriver.GetIbxQuery( const Handle : TOdbHandle ) : TIBQuery;
begin
    Result := GetHandleDataObj(Handle) as TIBQuery;
end;

{--------------------------------------------------------------------}

function TIbxOdbDriver.GetDataAccessMethods : TDataAccessMethodSet;
begin
    Result := [damSQL];
end;

{--------------------------------------------------------------------}

function TIbxOdbDriver.GetSupportedFieldTypes : TOdbFieldTypes;
begin
    Result := [oftString, oftInteger, oftFloat];
end;

{--------------------------------------------------------------------------}

function TIbxOdbDriver.AutoIncSupported              : Boolean;
begin
    Result := False;
end;

{--------------------------------------------------------------------------}

function TIbxOdbDriver.IntegerIndexSupported         : Boolean;
begin
    Result := True;
end;

{--------------------------------------------------------------------------}

function TIbxOdbDriver.FloatIndexSupported           : Boolean;
begin
    Result := True;
end;

{--------------------------------------------------------------------------}

function TIbxOdbDriver.CaseInsensitiveIndexSupported : Boolean;
begin
    Result := True;
end;

{--------------------------------------------------------------------------}

function TIbxOdbDriver.DescendingIndexSupported      : Boolean;
begin
    Result := True;
end;

{--------------------------------------------------------------------------}

function TIbxOdbDriver.TransactionsSupported         : Boolean;
begin
    Result := True;
end;

{--------------------------------------------------------------------------}

function TIbxOdbDriver.FileLockingSupported          : Boolean;
begin
    Result := False;
end;

{--------------------------------------------------------------------------}

function TIbxOdbDriver.ObjectLockingSupported        : Boolean;
begin
    Result := False;
end;

{--------------------------------------------------------------------------}

function TIbxOdbDriver.SingleIndexOnly               : Boolean;
begin
    Result := False;
end;

{--------------------------------------------------------------------}

{$ifdef __Testing}

procedure TIbxOdbDriverTest.GetParams(
                    var DriverClass                   : TOdbDriverClass;
                    var DatabaseName                  : String;
                    var ServerName                    : String;
                    var UserName                      : String;
                    var Password                      : String
                  );
begin
    DriverClass  := TIbxOdbDriver;
    DatabaseName := 'D:\P\Odb2\Tests\testdb\Interbase\ODB.GDB';
    ServerName   := '';
    UserName     := 'sysdba';
    Password     := 'masterkey';
end;

{$endif}

{--------------------------------------------------------------------}

begin
    RegisterOdbDriver(TIbxOdbDriver, 'Interbase Express');
end.
