{---------------------------------------------------------------------------
|
|  Module: OdbNav.Pas
|
|  Description: Object file navigator.
|
|  History: May 09, 1998. MB first version.
|           Mar 27, 2001. MB redesign to support TObjectQuery
|
|---------------------------------------------------------------------------}
unit OdbNav;

{$I Odb.Inc}

interface

uses
    {$ifdef __Testing}
    AutoTest, { for TSpiderOdbTestcase }
    TestObj,  { for TTestObjectMapping }
    {$endif}

    Buttons,  { for TSpeedButton }
    ExtCtrls, { for TCustomerPanel }
    StdCtrls, { for TComboBox }
    {$ifdef Win32}
    Windows,
    {$else}
    WinTypes,
    WinProcs,
    {$endif}
    Odb,      { for TObjectFile }

  SysUtils, Messages, Classes, Graphics, Controls,
  Forms, Dialogs;

type

{ Speed button with auto repeat when mouse is held down. }
TRepeatSpeedButton = class(TSpeedButton)
protected
    Timer : TTimer;
    procedure TimerExpired(Sender: TObject);
    procedure MouseDown (Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    procedure MouseUp (Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;

public
    constructor Create (AOwner : TComponent); override;
    destructor  Destroy; override;
end;


type

EOdbNavigatorEventNotAssigned = class(Exception);

TObjectToFormEvent = procedure( const Obj : TObject) of object;
TFormToObjectEvent = procedure( const Obj : TObject ) of object;
TNewObjectEvent    = procedure( var   Obj : TObject ) of object;

TMessageDlgFunc = function( const Msg     : String;
                                  DlgType : TMsgDlgType;
                                  Buttons : TMsgDlgButtons;
                                  HelpCtx : LongInt ): Integer;

TObjectDatasetNavigator = class(TCustomPanel)
protected
  FNavigatorPanel     : TPanel;
  FUpdateBtn          : TSpeedButton;
  FInsertBtn          : TSpeedButton;
  FDeleteBtn          : TSpeedButton;

  FFirstBtn           : TSpeedButton;
  FLastBtn            : TSpeedButton;
  FNextBtn            : TRepeatSpeedButton;
  FPriorBtn           : TRepeatSpeedButton;

  FBeforeUpdate       : TNotifyEvent;
  FAfterUpdate        : TNotifyEvent;

  FBeforeInsert       : TNotifyEvent;
  FAfterInsert        : TNotifyEvent;

  FBeforeDelete       : TNotifyEvent;
  FAfterDelete        : TNotifyEvent;

  FDataset            : TObjectDataset;

  FOnObjectToForm     : TObjectToFormEvent;
  FOnFormToObject     : TFormToObjectEvent;
  FOnNewObject        : TNewObjectEvent;

  FAutoPost           : Boolean;

  FEOFMessage            : String;
  FBOFMessage            : String;

  FEmptyMessage          : String;
  FConfirmDeleteMessage  : String;
  FConfirmChangeMessage  : String;
  FObjectNotFoundMessage : String;

  FMessageDlgFunc        : TMessageDlgFunc;

  procedure SetDataset(const InDataset : TObjectDataset); virtual;

  procedure CheckAssignments;

public
  { Public declarations }

  constructor Create( AOwner : TComponent ); override;

  destructor Destroy; override;

  procedure RefreshCurrent;

  { Checks if the current object displayed on the form has changed. If the
    object has changed, the object is saved automatically when AutoPost is True,
    otherwise the user is prompted to save the changed object. }
  procedure CheckDirty;

  property Dataset : TObjectDataset read FDataset
                                    write SetDataset;



  property UpdateBtn   : TSpeedButton read FUpdateBtn;
  property InsertBtn   : TSpeedButton read FInsertBtn;
  property DeleteBtn   :  TSpeedButton read FDeleteBtn;

  property FirstBtn      : TSpeedButton read FFirstBtn;
  property LastBtn       : TSpeedButton read FLastBtn;
  property NextBtn       : TRepeatSpeedButton read FNextBtn;
  property PriorBtn      : TRepeatSpeedButton read FPriorBtn;

published

  property AutoPost : Boolean read FAutoPost
                              write FAutoPost
                              default False;

  property OnNewObject : TNewObjectEvent
                read FOnNewObject
                write FOnNewObject;

  property OnObjectToForm : TObjectToFormEvent
                read FOnObjectToForm
                write FOnObjectToForm;

  property OnFormToObject : TFormToObjectEvent
                read FOnFormToObject
                write FOnFormToObject;

  property BeforeUpdate  : TNotifyEvent read FBeforeUpdate
                                        write FBeforeUpdate;
  property AfterUpdate   : TNotifyEvent read FAfterUpdate
                                        write FAfterUpdate;

  property BeforeInsert  : TNotifyEvent read FBeforeInsert
                                        write FBeforeInsert;
  property AfterInsert   : TNotifyEvent read FAfterInsert
                                        write FAfterInsert;

  property BeforeDelete  : TNotifyEvent read FBeforeDelete
                                        write FBeforeDelete;
  property AfterDelete   : TNotifyEvent read FAfterDelete
                                        write FAfterDelete;

protected

    procedure FirstClick( Sender : TObject );
    procedure NextClick( Sender : TObject );
    procedure LastClick( Sender : TObject );
    procedure PriorClick( Sender : TObject );

    procedure InsertClick( Sender : TObject ); virtual; abstract;
    procedure UpdateClick( Sender : TObject ); virtual; abstract;
    procedure DeleteClick( Sender : TObject ); virtual; abstract;
end;

TObjectFileNavigator = class( TObjectDatasetNavigator )
protected
    FFindPanel               : TPanel;
    FFindBtn                 : TSpeedButton;
    FIndexComboBox           : TComboBox;

    FSearchIndices           : TStrings;

    FShiftPressedOnFindClick : Boolean;

    procedure SetSearchIndices(const InSearchIndices : TStrings);

    procedure FindClick( Sender : TObject );

    procedure FindBtnMouseDown( Sender: TObject;
                                Button: TMouseButton;
                                Shift: TShiftState; X, Y: Integer);

    procedure IndexChange( Sender : TObject );

    procedure UpdateSearchIndices;

    procedure FormResize(Sender: TObject);

    procedure ResizeIndexComboBox;

    procedure SetDataset(const InDataset : TObjectDataset); override;

    procedure InsertClick( Sender : TObject ); override;
    procedure UpdateClick( Sender : TObject ); override;
    procedure DeleteClick( Sender : TObject ); override;

public

    constructor Create( AOwner : TComponent ); override;
    destructor Destroy; override;

    property FindBtn       : TSpeedButton read FFindBtn;
    property IndexComboBox : TComboBox read FIndexComboBox;

    procedure SetMessages( const EOFMessage            : String;
                           const BOFMessage            : String;
                           const EmptyMessage          : String;
                           const ConfirmDeleteMessage  : String;
                           const ConfirmChangeMessage  : String;
                           const ObjectNotFoundMessage : String;
                           const FirstHint             : String;
                           const NextHint              : String;
                           const LastHint              : String;
                           const PriorHint             : String;
                           const FindHint              : String;
                           const InsertHint            : String;
                           const UpdateHint            : String;
                           const DeleteHint            : String;
                           const SearchIndexHint       : String );

published

  property SearchIndices : TStrings read FSearchIndices
                                    write SetSearchIndices;

end;

{$ifdef __Testing}

type

TDatasetNavigatorTest = class( TSpiderOdbTestCase )
protected
     FObjectDatabase : TObjectDatabase;
     FDataset        : TObjectDataset;
     FNavigator      : TObjectDatasetNavigator;
     FParentForm     : TForm;

     { Object which simalates variables held by the form }
     FFormObj        : TTestObject;

     function CreateAndPopulateDataset : TObjectDataset; virtual; abstract;
     function CreateNavigator : TObjectDatasetNavigator; virtual; abstract;

     procedure InitTest;
     procedure TestDone;

     procedure OnObjectToForm( const Obj : TObject);
     procedure OnFormToObject( const Obj : TObject );
     procedure OnNewObject( var   Obj : TObject );

     procedure TestScanByName;
     procedure TestScanByWeight;

end;

TObjectFileNavigatorTest = class( TDatasetNavigatorTest )
protected
     function CreateAndPopulateDataset : TObjectDataset; override;
     function CreateNavigator : TObjectDatasetNavigator; override;

     function GetObjectFile : TObjectFile;
     property ObjectFile : TObjectFile read GetObjectFile;


     function GetNavigator : TObjectFileNavigator;
     property Navigator : TObjectFileNavigator read GetNavigator;

     procedure TestIndexComboList;
     procedure TestFindByName;
     procedure TestFindByWeight;
     procedure TestFindByAgeName;

published
     procedure Test;
end;


{$endif}

{--------------------------------------------------------------------}

implementation

uses
    {$ifdef __Testing}
    OdbBde,   { for TBdeOdbDriver }
    {$endif}
    ObjPerst, { for ObjectsEqual }
    OdbMisc,  { for WordCount, ExtractWord, LoadBitmapFromExeRes }
    OdbIndex, { for TOdbIndex }
    OdbField; { for FieldListHasAutoInc }

{$R OdbNav.R32}

const

{ bitmap names in OdbNav.RXX }
CUpdateBtnResName  = 'UPDATEBMP';
CInsertBtnResName  = 'ADDBMP';
CDeleteBtnResName  = 'DELETEBMP';
CFirstBtnResName   = 'FIRSTBMP';
CLastBtnResName    = 'LASTBMP';
CNextBtnResName    = 'NEXTBMP';
CPriorBtnResName   = 'PRIORBMP';
CFindBtnResName    = 'FINDBMP';

CButtonWidth       = 27;
CButtonHeight      = 25;

{---------------------------------------------------------------------------}

constructor TRepeatSpeedButton.Create (AOwner : TComponent);

begin
    inherited Create (AOwner);
end;

{---------------------------------------------------------------------------}

destructor TRepeatSpeedButton.Destroy;
begin
    Timer.Free;

    inherited Destroy;
end;

{---------------------------------------------------------------------------}

procedure TRepeatSpeedButton.MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
    inherited MouseDown (Button, Shift, X, Y);

    begin
        if not Assigned(Timer) then
        begin
            Timer          := TTimer.Create(Self);
            with Timer do
            begin
                OnTimer        := TimerExpired;
                Interval       := 500;
                Enabled        := True;
            end;
        end;
    end;
end;

{---------------------------------------------------------------------------}

procedure TRepeatSpeedButton.MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);

begin
    inherited MouseUp (Button, Shift, X, Y);
    Timer.Free;
    Timer := nil;
end;

{---------------------------------------------------------------------------}

{$ifdef Win32}
    {$J+}  { writable typed constants }
{$endif}

procedure TRepeatSpeedButton.TimerExpired(Sender: TObject);
const
    InTimerExpired : Boolean = False;
begin
    { prevent recursive calls. }
    if InTimerExpired then
        Exit;

    InTimerExpired := True;
    try
        if (fState = bsDown) and MouseCapture then
        begin
            Click;

            if Assigned(Timer) then
                Timer.Interval := 10
            else
                { ensure button is refreshed, is the mouse button has been
                  released within the timer event. }
                Refresh;
        end;
    finally
        InTimerExpired := False;
    end;
end;

{---------------------------------------------------------------------------}

constructor TObjectDatasetNavigator.Create( AOwner : TComponent );
var
    Bitmap : TBitmap;

begin
    inherited Create(AOwner);

    Left           := 0;
    Top            := 0;
    Width          := 632;
    Height         := 41;
    Align          := alTop;
    BevelOuter     := bvNone;
    Font.Color     := clBlack;
    Font.Height    := -15;
    Font.Name      := 'Arial';
    Font.Style     := [];

    FMessageDlgFunc := Dialogs.MessageDlg;

    { navigator panel }
    FNavigatorPanel := TPanel.Create(Self);
    with FNavigatorPanel do
    begin
        Parent     := Self;
        Left       := 0;
        Top        := 0;
        Width      := 218;
        Height     := 41;
        Align      := alLeft;
        BevelOuter := bvLowered;

        FFirstBtn := TSpeedButton.Create(FNavigatorPanel);
        with FFirstBtn do
        begin
            Parent    := FNavigatorPanel;
            Left      := 6;
            Top       := 8;
            Width     := CButtonWidth;
            Height    := CButtonHeight;
            Bitmap    := LoadBitmapFromExeRes( CFirstBtnResName );
            try
                Glyph := Bitmap;
            finally
                Bitmap.Free;
            end;

            NumGlyphs := 2;
            ShowHint  := True;
        end;

        FPriorBtn := TRepeatSpeedButton.Create(FNavigatorPanel);
        with FPriorBtn do
        begin
            Parent    := FNavigatorPanel;
            Left      := 32;
            Top       := 8;
            Width     := CButtonWidth;
            Height    := CButtonHeight;
            Bitmap    := LoadBitmapFromExeRes( CPriorBtnResName );
            try
                Glyph := Bitmap;
            finally
                Bitmap.Free;
            end;
            NumGlyphs := 2;
            ShowHint  := True;
        end;

        FNextBtn := TRepeatSpeedButton.Create(FNavigatorPanel);
        with FNextBtn do
        begin
            Parent    := FNavigatorPanel;
            Left      := 58;
            Top       := 8;
            Width     := CButtonWidth;
            Height    := CButtonHeight;
            Bitmap    := LoadBitmapFromExeRes( CNextBtnResName );
            try
                Glyph := Bitmap;
            finally
                Bitmap.Free;
            end;
            NumGlyphs := 2;
            ShowHint  := True;
        end;

        FLastBtn := TSpeedButton.Create(FNavigatorPanel);
        with FLastBtn do
        begin
            Parent    := FNavigatorPanel;
            Left      := 84;
            Top       := 8;
            Width     := CButtonWidth;
            Height    := CButtonHeight;
            Bitmap    := LoadBitmapFromExeRes( CLastBtnResName );
            try
                Glyph := Bitmap;
            finally
                Bitmap.Free;
            end;
            NumGlyphs := 2;
            ShowHint  := True;
        end;

        FUpdateBtn := TSpeedButton.Create(FNavigatorPanel);
        with FUpdateBtn do
        begin
            Parent    := FNavigatorPanel;
            Left      := 118;
            Top       := 8;
            Width     := CButtonWidth;
            Height    := CButtonHeight;
            Bitmap    := LoadBitmapFromExeRes( CUpdateBtnResName );
            try
                Glyph := Bitmap;
            finally
                Bitmap.Free;
            end;
            NumGlyphs := 2;
            ShowHint  := True;
        end;

        FInsertBtn := TSpeedButton.Create(FNavigatorPanel);
        with FInsertBtn do
        begin
            Parent    := FNavigatorPanel;
            Left      := 152;
            Top       := 8;
            Width     := CButtonWidth;
            Height    := CButtonHeight;
            Bitmap    := LoadBitmapFromExeRes( CInsertBtnResName );
            try
                Glyph := Bitmap;
            finally
                Bitmap.Free;
            end;
            NumGlyphs := 2;
            ShowHint  := True;
        end;

        FDeleteBtn := TSpeedButton.Create(FNavigatorPanel);
        with FDeleteBtn do
        begin
            Parent    := FNavigatorPanel;
            Left      := 186;
            Top       := 8;
            Width     := CButtonWidth;
            Height    := CButtonHeight;
            Bitmap    := LoadBitmapFromExeRes( CDeleteBtnResName );
            try
                Glyph := Bitmap;
            finally
                Bitmap.Free;
            end;
            NumGlyphs := 2;
            ShowHint  := True;
        end;
    end; { with FNavigatorPanel }

    FAutoPost := False;

    FFirstBtn.OnClick  := FirstClick;
    FNextBtn.OnClick   := NextClick;
    FLastBtn.OnClick   := LastClick;
    FPriorBtn.OnClick  := PriorClick;

    FInsertBtn.OnClick := InsertClick;
    FUpdateBtn.OnClick := UpdateClick;
    FDeleteBtn.OnClick := DeleteClick;
end;

{---------------------------------------------------------------------------}

destructor TObjectDatasetNavigator.Destroy;
begin
    { Setting Dataset will call CheckDirty }
    Dataset := nil;

    inherited Destroy;
end;

{---------------------------------------------------------------------------}

procedure TObjectDatasetNavigator.CheckAssignments;
begin
    if not Assigned(FDataset) then
        raise EOdbNavigatorEventNotAssigned.Create('Dataset not assigned');

    if not Assigned(FOnNewObject) then
        raise EOdbNavigatorEventNotAssigned.Create('OnNewObject not assigned');

    if not Assigned(FOnObjectToForm) then
        raise EOdbNavigatorEventNotAssigned.Create('OnObjectToForm not assigned');

    if not Assigned(FOnFormToObject) then
        raise EOdbNavigatorEventNotAssigned.Create('OnObjectToForm not assigned');
end;

{---------------------------------------------------------------------------}

procedure TObjectDatasetNavigator.CheckDirty;
var
    Obj : TObject;
begin
    if not FDataset.IsEmpty then
    begin
        Obj := CloneObj(FDataset.CurrentObj);
        try
            FOnFormToObject(Obj);

            if not ObjectsEqual(Obj, FDataset.CurrentObj) then
            begin
                if AutoPost then
                    UpdateClick(Self)
                else
                begin
                    if FMessageDlgFunc( FConfirmChangeMessage, mtConfirmation,
                                        [mbYes,mbNo], 0 ) = mrYes then
                        UpdateClick(Self);
                end;
            end;
        finally
            Obj.Free;
        end;
    end;
end;

{---------------------------------------------------------------------------}

procedure TObjectDatasetNavigator.RefreshCurrent;
begin
    CheckAssignments;

    if FDataset.IsEmpty then
    begin
        FMessageDlgFunc(FEmptyMessage, mtInformation, [mbOk], 0);
        Exit;
    end;

    FDataset.ReloadCurrent;
    FOnObjectToForm(FDataset.CurrentObj);
end;

{---------------------------------------------------------------------------}

procedure TObjectDatasetNavigator.SetDataset(const InDataset : TObjectDataset);
begin
    if FDataset <> nil then
        CheckDirty;

    FDataset := InDataset;

    if Assigned(FDataset) then
    begin
        if not FDataset.Opened then
            FDataset.Open;

        if FDataset.FindFirst then
            FOnObjectToForm(FDataset.CurrentObj)
    end;
end;

{---------------------------------------------------------------------------}

procedure TObjectDatasetNavigator.FirstClick( Sender : TObject );
begin
    CheckAssignments;

    CheckDirty;

    if FDataset.FindFirst then
        FOnObjectToForm(FDataset.CurrentObj)
    else
        FMessageDlgFunc(FEmptyMessage, mtInformation, [mbOk], 0);
end;

{---------------------------------------------------------------------------}

procedure TObjectDatasetNavigator.NextClick( Sender : TObject );
begin
    CheckAssignments;

    CheckDirty;

    if FDataset.FindNext then
        FOnObjectToForm(FDataset.CurrentObj)
    else
        FMessageDlgFunc(FEOFMessage, mtInformation, [mbOk], 0);
end;

{---------------------------------------------------------------------------}

procedure TObjectDatasetNavigator.LastClick( Sender : TObject );
begin
    CheckAssignments;

    CheckDirty;

    if FDataset.FindLast then
        FOnObjectToForm(FDataset.CurrentObj)
    else
        FMessageDlgFunc(FEmptyMessage, mtInformation, [mbOk], 0);
end;

{---------------------------------------------------------------------------}

procedure TObjectDatasetNavigator.PriorClick( Sender : TObject );
begin
    CheckAssignments;

    CheckDirty;

    if FDataset.FindPrior then
        FOnObjectToForm(FDataset.CurrentObj)
    else
        FMessageDlgFunc(FBOFMessage, mtInformation, [mbOk], 0);
end;

{---------------------------------------------------------------------------}

constructor TObjectFileNavigator.Create( AOwner : TComponent );
var
    Bitmap : TBitmap;
begin
    inherited Create(AOwner);

    FSearchIndices := TStringList.Create;
    FSearchIndices.Clear;

    { Find panel }
    FFindPanel := TPanel.Create(Self);
    with FFindPanel do
    begin
        Parent     := Self;
        Left       := 211;
        Top        := 0;
        Width      := 345;
        Height     := 41;
        Align      := alClient;
        BevelOuter := bvLowered;

        FFindBtn := TSpeedButton.Create(FFindPanel);
        with FFindBtn do
        begin
            Parent      := FFindPanel;
            Left        := 5;
            Top         := 8;
            Width       := CButtonWidth;
            Height      := CButtonHeight;
            OnMouseDown := FindBtnMouseDown;
            OnClick     := FindClick;
            Bitmap      := LoadBitmapFromExeRes( CFindBtnResName );
            try
                Glyph := Bitmap;
            finally
                Bitmap.Free;
            end;
            NumGlyphs := 2;
            ShowHint  := True;
        end;

        FIndexComboBox := TComboBox.Create(FFindPanel);
        with FIndexComboBox do
        begin
            Parent     := FFindPanel;
            Left       := 37;
            Top        := 8;
            Width      := 300;
            Height     := 25;
            ItemHeight := 17;
            ShowHint   := True;
            Font.Size  := 9;
            Style      := csDropDownList;
        end;
    end; { with FFindPanel }

    FIndexComboBox.OnChange := IndexChange;

    SetMessages( 'End of file is reached.',
                 'Beginning of file is reached.',
                 'File is empty.',
                 'Delete the current object ?',
                 'Current object has changed. Update ?',
                 'Object NOT found.',
                 'First',
                 'Next',
                 'Last',
                 'Prior',
                 'Find',
                 'Insert',
                 'Update',
                 'Delete',
                 'Search index' );


    Self.OnResize  := FormResize;
end;

{---------------------------------------------------------------------------}

destructor TObjectFileNavigator.Destroy;
begin
    FSearchIndices.Free;

    inherited Destroy;
end;

{---------------------------------------------------------------------------}

procedure TObjectFileNavigator.SetMessages(
                       const EOFMessage            : String;
                       const BOFMessage            : String;
                       const EmptyMessage          : String;
                       const ConfirmDeleteMessage  : String;
                       const ConfirmChangeMessage  : String;
                       const ObjectNotFoundMessage : String;
                       const FirstHint             : String;
                       const NextHint              : String;
                       const LastHint              : String;
                       const PriorHint             : String;
                       const FindHint              : String;
                       const InsertHint            : String;
                       const UpdateHint            : String;
                       const DeleteHint            : String;
                       const SearchIndexHint       : String );
begin
    FEOFMessage            := EofMessage;
    FBOFMessage            := BofMessage;
    FEmptyMessage          := EmptyMessage;
    FConfirmDeleteMessage  := ConfirmDeleteMessage;
    FConfirmChangeMessage  := ConfirmChangeMessage;
    FObjectNotFoundMessage := ObjectNotFoundMessage;

    FFirstBtn.Hint         := FirstHint;
    FNextBtn.Hint          := NextHint;
    FLastBtn.Hint          := LastHint;
    FPriorBtn.Hint         := PriorHint;
    FFindBtn.Hint          := FindHint;
    FInsertBtn.Hint        := InsertHint;
    FUpdateBtn.Hint        := UpdateHint;
    FDeleteBtn.Hint        := DeleteHint;
    FIndexComboBox.Hint    := SearchIndexHint;
end;

{---------------------------------------------------------------------------}

procedure TObjectFileNavigator.SetSearchIndices(const InSearchIndices : TStrings);
begin
    FSearchIndices.Assign(InSearchIndices);
end;

{---------------------------------------------------------------------------}

procedure TObjectFileNavigator.SetDataset(const InDataset : TObjectDataset);
begin
    inherited SetDataset(InDataset);

    if InDataset <> nil then
    begin
        UpdateSearchIndices;

        { invoke IndexChange to ensure the search index is initialy set.}
        IndexChange(Self);
    end;
end;

{---------------------------------------------------------------------------}

procedure TObjectFileNavigator.UpdateSearchIndices;
var
    Index   : TOdbIndex;
    IndexNo : Integer;
begin
    CheckAssignments;

    if FSearchIndices.Count = 0 then
    begin
        { if the user has not specified any search indices, use all
          indices defined in the object file. }
        if FDataset.Indices.GotoFirst then
        repeat
             Index := FDataset.Indices.CurrentObj as TOdbIndex;

             FSearchIndices.Add(Index.Name);
        until not FDataset.Indices.GotoNext;
    end;

    FIndexComboBox.Clear;
    for IndexNo := 0 to (FSearchIndices.Count-1) do
    begin
        if WordCount(FSearchIndices[IndexNo],[';']) = 2 then
            FIndexComboBox.Items.Add(ExtractWord(2, FSearchIndices[IndexNo], [';']))
        else
            FIndexComboBox.Items.Add(ExtractWord(1, FSearchIndices[IndexNo], [';']));
    end;

    FIndexComboBox.ItemIndex := 0;
end;

{---------------------------------------------------------------------------}

procedure TObjectFileNavigator.IndexChange( Sender : TObject );
begin
    (FDataset as TObjectFile).IndexName :=
        ExtractWord(1, FSearchIndices.Strings[FIndexComboBox.ItemIndex], [';']);
end;

{---------------------------------------------------------------------------}

procedure TObjectFileNavigator.FormResize(Sender: TObject);
begin
    ResizeIndexComboBox;
end;

{---------------------------------------------------------------------------}

procedure TObjectFileNavigator.ResizeIndexComboBox;
begin
    FIndexComboBox.Width := FFindPanel.Width - FIndexComboBox.Left - 8;

    if FIndexComboBox.Width < 100 then
        FIndexComboBox.Width := 100;
end;

{---------------------------------------------------------------------------}

procedure TObjectFileNavigator.InsertClick( Sender : TObject );
var
    Obj : TObject;
begin
    CheckAssignments;

    CheckDirty;

    if Assigned(FBeforeInsert) then
        FBeforeInsert(Self);

    FOnNewObject(Obj);
    try
        (FDataset as TObjectFile).CurrentObj := Obj;
    finally
        { MB Apr 06, 2001. Fix memory leak }
        Obj.Free;
    end;
    (FDataset as TObjectFile).Insert;
    FOnObjectToForm(FDataset.CurrentObj);

    if Assigned(FAfterInsert) then
        FAfterInsert(Self);
end;

{---------------------------------------------------------------------------}

procedure TObjectFileNavigator.UpdateClick( Sender : TObject );
var
    Obj : TObject;
begin
    CheckAssignments;

    if FDataset.IsEmpty then
    begin
        FMessageDlgFunc(FEmptyMessage, mtInformation, [mbOk], 0);
        Exit;
    end;

    if Assigned(FBeforeUpdate) then
        FBeforeUpdate(Self);

    Obj := FDataset.CurrentObj;
    FOnFormToObject(Obj);
    (FDataset as TObjectFile).Update;

    if Assigned(FAfterUpdate) then
        FAfterUpdate(Self);
end;

{---------------------------------------------------------------------------}

procedure TObjectFileNavigator.DeleteClick( Sender : TObject );
begin
    CheckAssignments;

    if FDataset.IsEmpty then
    begin
        FMessageDlgFunc(FEmptyMessage, mtInformation, [mbOk], 0);
        Exit;
    end;

    if (FMessageDlgFunc( FConfirmDeleteMessage, mtConfirmation,
                    [mbYes,mbNo], 0 ) <> mrYes) then
        Exit;

    if Assigned(FBeforeDelete) then
        FBeforeDelete(Self);

    (FDataset as TObjectFile).Delete;
    if not FDataset.IsEmpty then
        FOnObjectToForm(FDataset.CurrentObj);

    if Assigned(FAfterDelete) then
        FAfterDelete(Self);
end;

{---------------------------------------------------------------------------}

procedure TObjectFileNavigator.FindClick( Sender : TObject );
var
    Obj : TObject;
begin                                         
    CheckAssignments;

    if FDataset.IsEmpty then
    begin
        FMessageDlgFunc(FEmptyMessage, mtInformation, [mbOk], 0);
        Exit;
    end;

    Obj := FDataset.CurrentObj;
    FOnFormToObject(Obj);

    if FShiftPressedOnFindClick then
    begin
        if not (FDataset as TObjectFile).FindExact then
            FMessageDlgFunc(FObjectNotFoundMessage, mtInformation, [mbOk], 0)
    end
    else
    begin
        (FDataset as TObjectFile).FindNearest;
    end;

    FOnObjectToForm(FDataset.CurrentObj);
end;

{---------------------------------------------------------------------------}

procedure TObjectFileNavigator.FindBtnMouseDown(
                            Sender: TObject;
                            Button: TMouseButton;
                            Shift: TShiftState; X, Y: Integer);
begin
    if ssShift in Shift then
        FShiftPressedOnFindClick := True
    else
        FShiftPressedOnFindClick  := False;
end;

{---------------------------------------------------------------------------}

{$ifdef __Testing}

var

{ return value used by TestMessageDlg }
MessageDlgResult : Integer;
MessageDlgMsg    : String;

function TestMessageDlg( const Msg     : String;
                               DlgType : TMsgDlgType;
                               Buttons : TMsgDlgButtons;
                               HelpCtx : LongInt ): Integer;
begin
    MessageDlgMsg := Msg;
    Result        := MessageDlgResult;
end;

{---------------------------------------------------------------------------}

procedure TDatasetNavigatorTest.InitTest;
begin
    FFormObj          := TTestObject.Create;

    FObjectDatabase := TObjectDatabase.Create( TBdeOdbDriver, Self.TestDbPath );
    FObjectDatabase.Open( '', '', '' );

    FDataset          := CreateAndPopulateDataset;
    FParentForm       := TForm.Create(nil);
    FNavigator        := CreateNavigator;
    FNavigator.Parent := FParentForm;

    FNavigator.FMessageDlgFunc := TestMessageDlg;

    FNavigator.OnFormToObject  := Self.OnFormToObject;
    FNavigator.OnObjectToForm  := Self.OnObjectToForm;
    FNavigator.OnNewObject     := Self.OnNewObject;

    FNavigator.Dataset         := FDataset;
end;

{---------------------------------------------------------------------------}

procedure TDatasetNavigatorTest.TestDone;
begin
    FNavigator.Free;

    FParentForm.Free;

    FDataset.Close;
    FDataset.DeleteDataset;
    FDataset.Free;

    FObjectDatabase.Close;
    FObjectDatabase.Free;

    FFormObj.Free;
end;

{---------------------------------------------------------------------------}

procedure TDatasetNavigatorTest.OnObjectToForm( const Obj : TObject);
begin
    FFormObj.Name     := (Obj as TTestObject).Name;
    FFormObj.Age      := (Obj as TTestObject).Age;
    FFormObj.Id       := (Obj as TTestObject).Id;
    FFormObj.Weight   := (Obj as TTestObject).Weight;

    FFormObj.KeyValue := (Obj as TTestObject).KeyValue;

    FFormObj.Data     := (Obj as TTestObject).Data;

    FFormObj.Stream.Clear;
    FFormObj.Stream.CopyFrom((Obj as TTestObject).Stream, 0);
end;

{---------------------------------------------------------------------------}

procedure TDatasetNavigatorTest.OnFormToObject( const Obj : TObject );
begin
    (Obj as TTestObject).Name     := FFormObj.Name;
    (Obj as TTestObject).Age      := FFormObj.Age;
    (Obj as TTestObject).Id       := FFormObj.Id;
    (Obj as TTestObject).Weight   := FFormObj.Weight;
    (Obj as TTestObject).KeyValue := FFormObj.KeyValue;
    (Obj as TTestObject).Data     := FFormObj.Data;

    (Obj as TTestObject).Stream.Clear;
    (Obj as TTestObject).Stream.CopyFrom(FFormObj.Stream, 0);
end;

{---------------------------------------------------------------------------}

procedure TDatasetNavigatorTest.OnNewObject( var   Obj : TObject );
begin
    Obj := TTestObject.Create;
end;

{---------------------------------------------------------------------------}

procedure TDatasetNavigatorTest.TestScanByName;
begin
    { assert: the index of SQL property is properly set }

    FNavigator.FirstBtn.Click;
    Check( FFormObj.Name = 'Object1' );

    FNavigator.NextBtn.Click;
    Check( FFormObj.Name = 'Object2' );

    FNavigator.PriorBtn.Click;
    Check( FFormObj.Name = 'Object1' );

    FNavigator.LastBtn.Click;
    Check( FFormObj.Name = 'Object9' );

    FNavigator.PriorBtn.Click;
    Check( FFormObj.Name = 'Object8' );

    FNavigator.FirstBtn.Click;
    Check( FFormObj.Name = 'Object1' );

    FNavigator.NextBtn.Click;
    Check( FFormObj.Name = 'Object2' );

    FNavigator.NextBtn.Click;
    Check( FFormObj.Name = 'Object3' );

    FNavigator.NextBtn.Click;
    Check( FFormObj.Name = 'Object4' );

    FNavigator.NextBtn.Click;
    Check( FFormObj.Name = 'Object5' );

    FNavigator.NextBtn.Click;
    Check( FFormObj.Name = 'Object6' );

    FNavigator.NextBtn.Click;
    Check( FFormObj.Name = 'Object7' );

    FNavigator.NextBtn.Click;
    Check( FFormObj.Name = 'Object8' );

    FNavigator.NextBtn.Click;
    Check( FFormObj.Name = 'Object9' );

    FNavigator.LastBtn.Click;
    Check( FFormObj.Name = 'Object9' );

    FNavigator.PriorBtn.Click;
    Check( FFormObj.Name = 'Object8' );

    FNavigator.PriorBtn.Click;
    Check( FFormObj.Name = 'Object7' );

    FNavigator.PriorBtn.Click;
    Check( FFormObj.Name = 'Object6' );

    FNavigator.PriorBtn.Click;
    Check( FFormObj.Name = 'Object5' );

    FNavigator.PriorBtn.Click;
    Check( FFormObj.Name = 'Object4' );

    FNavigator.PriorBtn.Click;
    Check( FFormObj.Name = 'Object3' );

    FNavigator.PriorBtn.Click;
    Check( FFormObj.Name = 'Object2' );

    FNavigator.PriorBtn.Click;
    Check( FFormObj.Name = 'Object1' );

    FNavigator.PriorBtn.Click;
    Check( FFormObj.Name = 'Object1' );
end;

{---------------------------------------------------------------------------}

procedure TDatasetNavigatorTest.TestScanByWeight;
begin
    FNavigator.FirstBtn.Click;
    Check( FFormObj.Name = 'Object9' );

    FNavigator.NextBtn.Click;
    Check( FFormObj.Name = 'Object8' );

    FNavigator.PriorBtn.Click;
    Check( FFormObj.Name = 'Object9' );

    FNavigator.LastBtn.Click;
    Check( FFormObj.Name = 'Object1' );

    FNavigator.PriorBtn.Click;
    Check( FFormObj.Name = 'Object2' );

    FNavigator.FirstBtn.Click;
    Check( FFormObj.Name = 'Object9' );

    FNavigator.NextBtn.Click;
    Check( FFormObj.Name = 'Object8' );

    FNavigator.NextBtn.Click;
    Check( FFormObj.Name = 'Object7' );

    FNavigator.NextBtn.Click;
    Check( FFormObj.Name = 'Object6' );

    FNavigator.NextBtn.Click;
    Check( FFormObj.Name = 'Object5' );

    FNavigator.NextBtn.Click;
    Check( FFormObj.Name = 'Object4' );

    FNavigator.NextBtn.Click;
    Check( FFormObj.Name = 'Object3' );

    FNavigator.NextBtn.Click;
    Check( FFormObj.Name = 'Object2' );

    FNavigator.NextBtn.Click;
    Check( FFormObj.Name = 'Object1' );

    FNavigator.LastBtn.Click;
    Check( FFormObj.Name = 'Object1' );

    FNavigator.PriorBtn.Click;
    Check( FFormObj.Name = 'Object2' );

    FNavigator.PriorBtn.Click;
    Check( FFormObj.Name = 'Object3' );

    FNavigator.PriorBtn.Click;
    Check( FFormObj.Name = 'Object4' );

    FNavigator.PriorBtn.Click;
    Check( FFormObj.Name = 'Object5' );

    FNavigator.PriorBtn.Click;
    Check( FFormObj.Name = 'Object6' );

    FNavigator.PriorBtn.Click;
    Check( FFormObj.Name = 'Object7' );

    FNavigator.PriorBtn.Click;
    Check( FFormObj.Name = 'Object8' );

    FNavigator.PriorBtn.Click;
    Check( FFormObj.Name = 'Object9' );

    FNavigator.PriorBtn.Click;
    Check( FFormObj.Name = 'Object9' );
end;

{---------------------------------------------------------------------------}

function TObjectFileNavigatorTest.CreateAndPopulateDataset : TObjectDataset;
begin
    Result := TObjectFile.Create( FObjectDatabase, 'TestData', TTestObjectMapping );

    Result.CreateDataset;

    Result.Open;

    AddObjectISAM( 1, TObjectFile(Result) );
    AddObjectISAM( 2, TObjectFile(Result) );
    AddObjectISAM( 3, TObjectFile(Result) );
    AddObjectISAM( 4, TObjectFile(Result) );
    AddObjectISAM( 5, TObjectFile(Result) );
    AddObjectISAM( 6, TObjectFile(Result) );
    AddObjectISAM( 7, TObjectFile(Result) );
    AddObjectISAM( 8, TObjectFile(Result) );
    AddObjectISAM( 9, TObjectFile(Result) );
end;

{---------------------------------------------------------------------------}

function TObjectFileNavigatorTest.CreateNavigator : TObjectDatasetNavigator;
begin
    Result := TObjectFileNavigator.Create(nil);
end;

{---------------------------------------------------------------------------}

procedure TObjectFileNavigatorTest.Test;
var
    Count : Integer;
begin
    InitTest;
    try
        TestIndexComboList;

        { test first object positioned }
        Check( FFormObj.Name = 'Object1' );

        Navigator.FIndexComboBox.ItemIndex := 6;
        Navigator.FIndexComboBox.OnChange(Self);
        TestScanByWeight;

        Navigator.FIndexComboBox.ItemIndex := 0;
        Navigator.FIndexComboBox.OnChange(Self);
        TestScanByName;

        TestFindByName;

        TestFindByWeight;

        TestFindByAgeName;

        FNavigator.InsertBtn.Click;
        FFormObj.Name := 'Object95';
        FNavigator.UpdateBtn.Click;

        Navigator.FIndexComboBox.ItemIndex := 0;
        Navigator.FIndexComboBox.OnChange(Self);

        FNavigator.LastBtn.Click;
        Check( FFormObj.Name = 'Object95' );

        FNavigator.PriorBtn.Click;
        Check( FFormObj.Name = 'Object9' );

        { test refusing a deletion }
        FNavigator.LastBtn.Click;
        FNavigator.FMessageDlgFunc := TestMessageDlg;
        MessageDlgResult := mrNo;
        FNavigator.DeleteBtn.Click;
        Check( FFormObj.Name = 'Object95' );

        { test accepting a deletion }
        FNavigator.LastBtn.Click;
        FNavigator.FMessageDlgFunc := TestMessageDlg;
        MessageDlgResult := mrYes;
        FNavigator.DeleteBtn.Click;
        Check( FFormObj.Name = 'Object9' );

        { refuse an update when the current object is modifed }
        FNavigator.AutoPost := False;
        FNavigator.LastBtn.Click;
        Check( FFormObj.Name = 'Object9' );
        FFormObj.KeyValue := '345';
        MessageDlgResult := mrNo;
        FNavigator.FMessageDlgFunc := TestMessageDlg;
        FNavigator.PriorBtn.Click;
        Check( FFormObj.Name = 'Object8' );
        FNavigator.LastBtn.Click;
        Check( FFormObj.Name = 'Object9' );
        Check( FFormObj.KeyValue = '9' );

        { accept an update when the current object is modified }
        FNavigator.AutoPost := False;
        FNavigator.LastBtn.Click;
        Check( FFormObj.Name = 'Object9' );
        FFormObj.KeyValue := '345';
        MessageDlgResult := mrYes;
        FNavigator.FMessageDlgFunc := TestMessageDlg;
        FNavigator.PriorBtn.Click;
        Check( FFormObj.Name = 'Object8' );
        FNavigator.LastBtn.Click;
        Check( FFormObj.Name = 'Object9' );
        Check( FFormObj.KeyValue = '345' );

        { accept an update when the current object is modifed }
        FNavigator.AutoPost := True;
        FNavigator.LastBtn.Click;
        Check( FFormObj.Name = 'Object9' );
        FFormObj.KeyValue := '678';
        { set the message function to nil to ensure it is not called when
          AutoPost is True }
        FNavigator.FMessageDlgFunc := nil;
        FNavigator.PriorBtn.Click;
        Check( FFormObj.Name = 'Object8' );
        FNavigator.LastBtn.Click;
        Check( FFormObj.Name = 'Object9' );
        Check( FFormObj.KeyValue = '678' );

        FFormObj.Name     := 'Object32';
        FFormObj.KeyValue := '';
        FNavigator.RefreshCurrent;
        Check( FFormObj.Name = 'Object9' );
        Check( FFormObj.KeyValue = '678' );

        { test that CheckDirty is called when changing the dataset }
        FFormObj.KeyValue   := '9';
        FNavigator.AutoPost := False;
        FNavigator.FMessageDlgFunc := TestMessageDlg;
        MessageDlgResult    := mrYes;
        MessageDlgMsg       := '';
        FNavigator.Dataset  := nil;
        Check( MessageDlgMsg = FNavigator.FConfirmChangeMessage );

        FNavigator.Dataset := FDataset;

        { delete all records, to test on empty dataset }
        FNavigator.FMessageDlgFunc := TestMessageDlg;
        FNavigator.LastBtn.Click;
        MessageDlgResult := mrYes;
        for Count := 9 downto 2 do
        begin
            FNavigator.DeleteBtn.Click;
            Check( FFormObj.Name = 'Object' + IntToStr(Count-1) );
        end;

        { assert: only Object1 left }
        Check( FFormObj.Name = 'Object1');

        FNavigator.FMessageDlgFunc := TestMessageDlg;
        MessageDlgResult := mrYes;
        FNavigator.DeleteBtn.Click;

        { test operations on empty dataset }

        FNavigator.FMessageDlgFunc := TestMessageDlg;
        MessageDlgMsg := '';
        FNavigator.LastBtn.Click;
        Check( MessageDlgMsg = FNavigator.FEmptyMessage );

        MessageDlgMsg := '';
        FNavigator.NextBtn.Click;
        Check( MessageDlgMsg = FNavigator.FEOFMessage );

        MessageDlgMsg := '';
        FNavigator.FirstBtn.Click;
        Check( MessageDlgMsg = FNavigator.FEmptyMessage );

        MessageDlgMsg := '';
        FNavigator.PriorBtn.Click;
        Check( MessageDlgMsg = FNavigator.FBOFMessage );

        MessageDlgMsg := '';
        FNavigator.DeleteBtn.Click;
        Check( MessageDlgMsg = FNavigator.FEmptyMessage );

        MessageDlgMsg := '';
        FNavigator.UpdateBtn.Click;
        Check( MessageDlgMsg = FNavigator.FEmptyMessage );

        MessageDlgMsg := '';
        FNavigator.RefreshCurrent;
        Check( MessageDlgMsg = FNavigator.FEmptyMessage );

        MessageDlgMsg := '';
        Navigator.FindBtn.Click;
        Check( MessageDlgMsg = FNavigator.FEmptyMessage );


    finally
        TestDone;
    end;
end;

{---------------------------------------------------------------------------}

procedure TObjectFileNavigatorTest.TestIndexComboList;
begin
    Check( Navigator.FIndexComboBox.Items[0] = 'Name' );
    Check( Navigator.FIndexComboBox.Items[1] = 'KeyValue' );
    Check( Navigator.FIndexComboBox.Items[2] = 'NameCI' );
    Check( Navigator.FIndexComboBox.Items[3] = 'Age' );
    Check( Navigator.FIndexComboBox.Items[4] = 'AgeName' );
    Check( Navigator.FIndexComboBox.Items[5] = 'AgeDesc' );
    Check( Navigator.FIndexComboBox.Items[6] = 'Weight' );
    Check( Navigator.FIndexComboBox.Items[7] = 'WName' );
    Check( Navigator.FIndexComboBox.Items[8] = 'WAge' );
    Check( Navigator.FIndexComboBox.Items[9] = 'Id' );
end;

{---------------------------------------------------------------------------}

procedure TObjectFileNavigatorTest.TestFindByName;
begin
    Navigator.FIndexComboBox.ItemIndex := 0;
    Navigator.FIndexComboBox.OnChange(Self);
    Navigator.FirstBtn.Click;
    Check( FFormObj.Name = 'Object1' );

    { simulate without Shift key pressed, locate nearest }
    Navigator.FindBtnMouseDown( Self, mbLeft, [], 5, 5 );
    FFormObj.Name := 'Object35';
    Navigator.FindBtn.Click;
    Check( FFormObj.Name = 'Object4' );

    { simulate with Shift key pressed, locate exact fail }
    Navigator.FindBtnMouseDown( Self, mbLeft, [ssShift], 5, 5 );
    FFormObj.Name := 'Object35';
    MessageDlgMsg := '';
    FNavigator.FMessageDlgFunc := TestMessageDlg;
    Navigator.FindBtn.Click;
    Check( MessageDlgMsg = FNavigator.FObjectNotFoundMessage );

    Navigator.FIndexComboBox.Items[0] := 'Name';
    Navigator.FirstBtn.Click;
    Check( FFormObj.Name = 'Object1' );

    Navigator.FindBtnMouseDown( Self, mbLeft, [ssShift], 5, 5 );
    FFormObj.Name := 'Object4';
    Navigator.FindBtn.Click;
    Check( FFormObj.Name = 'Object4' );
end;

{---------------------------------------------------------------------------}

procedure TObjectFileNavigatorTest.TestFindByWeight;
begin
    Navigator.FIndexComboBox.ItemIndex := 6;
    Navigator.FIndexComboBox.OnChange(Self);
    Navigator.FirstBtn.Click;
    Check( FFormObj.Name = 'Object9' );

    (*
    { simulate without Shift key pressed, locate nearest }
    Navigator.FindBtnMouseDown( Self, mbLeft, [], 5, 5 );
    FFormObj.Name := 'Object35';
    Navigator.FindBtn.Click;
    Check( FFormObj.Name = 'Object4' );

    { simulate with Shift key pressed, locate exact fail }
    Navigator.FindBtnMouseDown( Self, mbLeft, [ssShift], 5, 5 );
    FFormObj.Name := 'Object35';
    MessageDlgMsg := '';
    FNavigator.FMessageDlgFunc := TestMessageDlg;
    Navigator.FindBtn.Click;
    Check( MessageDlgMsg = FNavigator.FObjectNotFoundMessage );

    Navigator.FIndexComboBox.Items[0] := 'Name';
    Navigator.FirstBtn.Click;
    Check( FFormObj.Name = 'Object1' );

    Navigator.FindBtnMouseDown( Self, mbLeft, [ssShift], 5, 5 );
    FFormObj.Name := 'Object4';
    Navigator.FindBtn.Click;
    Check( FFormObj.Name = 'Object4' );
    *)
end;

{---------------------------------------------------------------------------}

procedure TObjectFileNavigatorTest.TestFindByAgeName;
begin
end;

{---------------------------------------------------------------------------}

function TObjectFileNavigatorTest.GetNavigator : TObjectFileNavigator;
begin
    Result := FNavigator as TObjectFileNavigator;
end;

{---------------------------------------------------------------------------}

function TObjectFileNavigatorTest.GetObjectFile : TObjectFile;
begin
    Result := FDataset as TObjectFile;
end;

{---------------------------------------------------------------------------}

{$endif}

end.
